from typing import Optional, Any, Dict, List
from ..types import SpanType

import dataclasses


@dataclasses.dataclass
class BaseArguments:
    def to_kwargs(self, ignore_keys: Optional[List[str]] = None) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        ignore_keys = [] if ignore_keys is None else ignore_keys
        for key, value in self.__dict__.items():
            if (value is not None) and (key not in ignore_keys):
                result[key] = value

        return result


@dataclasses.dataclass
class EndSpanParameters(BaseArguments):
    """
    Span data parameters that we set (or might set) when the
    tracked function is ended.
    """

    metadata: Optional[Any] = None
    input: Optional[Dict[str, Any]] = None
    output: Optional[Dict[str, Any]] = None
    tags: Optional[List[str]] = None
    usage: Optional[Dict[str, Any]] = None


@dataclasses.dataclass
class StartSpanParameters(BaseArguments):
    """
    Span data parameters that we set (or might set) when the
    tracked function is started.
    """

    type: SpanType
    name: str
    tags: Optional[List[str]] = None
    metadata: Optional[Dict[str, Any]] = None
    input: Optional[Dict[str, Any]] = None



import getpass
import logging
from typing import Final, List, Optional

import httpx
import opik.config
from opik.config import (
    OPIK_BASE_URL_CLOUD,
    OPIK_BASE_URL_LOCAL,
    OPIK_WORKSPACE_DEFAULT_NAME,
)
from opik.configurator.interactive_helpers import ask_user_for_approval, is_interactive
from opik.exceptions import ConfigurationError

LOGGER = logging.getLogger(__name__)

HEALTH_CHECK_URL_POSTFIX: Final[str] = "/is-alive/ping"
HEALTH_CHECK_TIMEOUT: Final[float] = 1.0

URL_ACCOUNT_DETAILS_DEFAULT: Final[str] = (
    "https://www.comet.com/api/rest/v2/account-details"
)
URL_ACCOUNT_DETAILS_POSTFIX: Final[str] = "/rest/v2/account-details"

URL_WORKSPACE_GET_LIST_DEFAULT: Final[str] = (
    "https://www.comet.com/api/rest/v2/workspaces"
)
URL_WORKSPACE_GET_LIST_POSTFIX: Final[str] = "/rest/v2/workspaces"


class OpikConfigurator:
    def __init__(
        self,
        api_key: Optional[str] = None,
        workspace: Optional[str] = None,
        url: Optional[str] = None,
        use_local: bool = False,
        force: bool = False,
    ):
        self.api_key = api_key
        self.workspace = workspace
        self.url = url
        self.use_local = use_local
        self.force = force
        self.current_config = opik.config.OpikConfig()

    def configure(self) -> None:
        """
        Create a local configuration file for the Python SDK. If a configuration file already exists,
        it will not be overwritten unless the `force` parameter is set to True.

        Raises:
            ConfigurationError
            ConnectionError
        """

        # OPIK CLOUD
        if self.use_local is False:
            self._configure_cloud()
            return

        # LOCAL OPIK DEPLOYMENT
        self._configure_local()
        return

    def _configure_cloud(self) -> None:
        """
        Configure the cloud Opik instance by handling API key and workspace settings.
        """
        # Handle URL
        if self.url is None:
            self.url = OPIK_BASE_URL_CLOUD

        # Handle API key: get or prompt for one if needed
        update_config_with_api_key = self._set_api_key()

        # Handle workspace: get or prompt for one if needed
        update_config_with_workspace = self._set_workspace()

        # Update configuration if either API key or workspace has changed
        if update_config_with_api_key or update_config_with_workspace:
            self._update_config()
        else:
            self._update_config(save_to_file=False)
            LOGGER.info(
                "Opik is already configured. You can check the settings by viewing the config file at %s",
                self.current_config.config_file_fullpath,
            )

    def _configure_local(self) -> None:
        """
        Configure the local Opik instance by setting the local URL and workspace.

        Raises:
            ConfigurationError: Raised if the Opik instance is not active or not found.
        """
        self.api_key = None
        self.workspace = OPIK_WORKSPACE_DEFAULT_NAME

        # Step 1: If the URL is provided and active, update the configuration
        if self.url is not None and self._is_instance_active(self.url):
            self._update_config(save_to_file=self.force)
            return

        # Step 2: Check if the default local instance is active
        if self._is_instance_active(OPIK_BASE_URL_LOCAL):
            if (
                not self.force
                and self.current_config.url_override == OPIK_BASE_URL_LOCAL
            ):
                LOGGER.info(
                    f"Opik is already configured to local instance at {OPIK_BASE_URL_LOCAL}."
                )
                return

            # Step 3: Ask user if they want to use the found local instance
            if not is_interactive():
                raise ConfigurationError(
                    "Non-interactive mode detected. Unable to proceed."
                )

            use_url = ask_user_for_approval(
                f"Found local Opik instance on: {OPIK_BASE_URL_LOCAL}, do you want to use it? (Y/n)"
            )
            if use_url:
                self.url = OPIK_BASE_URL_LOCAL
                self._update_config()
                return

        # Step 4: Ask user for URL if no valid local instance is found or approved
        if not is_interactive():
            raise ConfigurationError(
                "Non-interactive mode detected. Unable to proceed as no local Opik instance was found."
            )
        self._ask_for_url()
        self._update_config()

    def _set_api_key(self) -> bool:
        """
        Determines and set the correct API key based on the current configuration, force flag, and user input.

        Returns:
            bool: a boolean indicating if the configuration file needs updating.
        """
        config_file_needs_updating = False

        if self.api_key:
            if not self._is_api_key_correct(self.api_key):
                raise ConfigurationError("API key is incorrect.")
            config_file_needs_updating = True if self.force else False

        elif self.force and self.api_key is None:
            self._ask_for_api_key()
            config_file_needs_updating = True

        elif self.api_key is None and self.current_config.api_key is None:
            self._ask_for_api_key()
            config_file_needs_updating = True

        elif self.api_key is None and self.current_config.api_key is not None:
            self.api_key = self.current_config.api_key

        return config_file_needs_updating

    def _ask_for_api_key(self) -> None:
        """
        Prompt the user for an Opik cloud API key and verify its validity.
        The function retries up to 3 times if the API key is invalid.

        Raises:
            ConfigurationError: Raised if the API key provided by the user is invalid after 3 attempts.
        """
        retries = 3

        LOGGER.info(
            "Your Opik cloud API key is available at https://www.comet.com/api/my/settings/."
        )

        if not is_interactive():
            raise ConfigurationError(
                "Non-interactive mode detected. Unable to proceed as no API key has been specified."
            )

        while retries > 0:
            user_input_api_key = getpass.getpass(
                "Please enter your Opik Cloud API key:"
            )

            if self._is_api_key_correct(user_input_api_key):
                self.api_key = user_input_api_key
                return
            else:
                LOGGER.error(
                    f"The API key provided is not valid on {OPIK_BASE_URL_CLOUD}. Please try again."
                )
                retries -= 1
        raise ConfigurationError("API key is incorrect.")

    def _is_api_key_correct(self, api_key: str) -> bool:
        """
        Validates if the provided Opik API key is correct by sending a request to the cloud API.

        Returns:
            bool: True if the API key is valid (status 200), False if the key is invalid (status 401 or 403).

        Raises:
            ConnectionError: If a network-related error occurs or the response status is neither 200, 401, nor 403.
        """

        try:
            with httpx.Client() as client:
                client.headers.update({"Authorization": f"{api_key}"})
                response = client.get(url=self._url_account_details)
            if response.status_code == 200:
                return True
            elif response.status_code in [401, 403]:
                return False
            else:
                raise ConnectionError(f"Error while checking API key: {response.text}")
        except httpx.RequestError as e:
            raise ConnectionError(f"Network error occurred: {str(e)}")
        except Exception as e:
            raise ConnectionError(f"Unexpected error occurred: {str(e)}")

    def _set_workspace(self) -> bool:
        """
        Determines and set the correct workspace based on current configuration, force flag, and user input.

        Returns:
            bool: a boolean indicating whether the configuration file needs updating.

        Raises:
            ConfigurationError: If the provided workspace is invalid.
        """

        # Case 1: Workspace was provided by the user and is valid
        if self.workspace is not None:
            if not self._is_workspace_name_correct(self.workspace):
                raise ConfigurationError(
                    f"Workspace `{self.workspace}` is incorrect for the given API key."
                )
            return True if self.force else False

        # Case 2: Use workspace from current configuration if not forced to change
        if (
            "workspace" in self.current_config.model_fields_set
            and self.current_config.workspace != OPIK_WORKSPACE_DEFAULT_NAME
            and not self.force
        ):
            self.workspace = self.current_config.workspace
            return False

        # Case 3: No workspace provided, prompt the user
        default_workspace = self._get_default_workspace()
        use_default_workspace = ask_user_for_approval(
            f'Do you want to use "{default_workspace}" workspace? (Y/n)'
        )

        if use_default_workspace:
            self.workspace = default_workspace
        else:
            self._ask_for_workspace()

        return True

    def _is_workspace_name_correct(self, workspace: str) -> bool:
        """
        Verifies whether the provided workspace name exists in the user's cloud Opik account.

        Args:
            workspace (str): The name of the workspace to check.

        Returns:
            bool: True if the workspace is found, False otherwise.

        Raises:
            ConnectionError: Raised if there's an issue with connecting to the Opik service, or the response is not successful.
        """
        if not self.api_key:
            raise ConfigurationError("API key must be set to check workspace name.")

        try:
            with httpx.Client() as client:
                client.headers.update({"Authorization": f"{self.api_key}"})
                response = client.get(url=self._url_get_workspace_list)
        except httpx.RequestError as e:
            # Raised for network-related errors such as timeouts
            raise ConnectionError(f"Network error: {str(e)}")
        except Exception as e:
            raise ConnectionError(f"Unexpected error occurred: {str(e)}")

        if response.status_code != 200:
            raise ConnectionError(
                f"HTTP error: {response.status_code} - {response.text}"
            )

        workspaces: List[str] = response.json().get("workspaceNames", [])
        return workspace in workspaces

    def _get_default_workspace(self) -> str:
        """
        Retrieves the default Opik workspace name associated with the given API key.

        Returns:
            str: The default workspace name.

        Raises:
            ConnectionError: If there's an error while fetching the default workspace.
        """
        if not self.api_key:
            raise ConfigurationError("API key must be set.")

        try:
            with httpx.Client() as client:
                client.headers.update({"Authorization": f"{self.api_key}"})
                response = client.get(url=self._url_account_details)

            if response.status_code != 200:
                raise ConnectionError(
                    f"Error while getting default workspace name: {response.text}"
                )

            default_workspace_name = response.json().get("defaultWorkspaceName")
            if not default_workspace_name:
                raise ConnectionError("defaultWorkspaceName not found in the response.")

            return default_workspace_name

        except httpx.RequestError as e:
            raise ConnectionError(f"Network error occurred: {str(e)}")
        except Exception as e:
            raise ConnectionError(f"Unexpected error occurred: {str(e)}")

    def _ask_for_workspace(self) -> None:
        """
        Prompt the user for an Opik instance workspace name and verify its validity.
        The function retries up to 3 times if the workspace name is invalid.

        Raises:
            ConfigurationError: Raised if the workspace name is invalid after 3 attempts.
        """
        retries = 3

        if not self.api_key:
            raise ConfigurationError("API key must be set to check workspace name.")

        if not is_interactive():
            raise ConfigurationError(
                "Non-interactive mode detected. Unable to proceed as no workspace name has been specified."
            )

        while retries > 0:
            user_input_workspace = input(
                "Please enter your cloud Opik instance workspace name: "
            )
            if self._is_workspace_name_correct(user_input_workspace):
                self.workspace = user_input_workspace
                return
            else:
                LOGGER.error(
                    "This workspace does not exist, please enter a workspace that you have access to."
                )
                retries -= 1
        raise ConfigurationError(
            "User does not have access to the workspaces provided."
        )

    def _update_config(self, save_to_file: bool = True) -> None:
        """
        Save changes to the config file and update the current session configuration.

        Raises:
            ConfigurationError: Raised if there is an issue saving the configuration or updating the session.
        """
        try:
            if save_to_file:
                new_config = opik.config.OpikConfig(
                    api_key=self.api_key,
                    url_override=self.url,
                    workspace=self.workspace,
                )
                new_config.save_to_file()

            # Update current session configuration
            opik.config.update_session_config("api_key", self.api_key)
            opik.config.update_session_config("url_override", self.url)
            opik.config.update_session_config("workspace", self.workspace)
        except Exception as e:
            LOGGER.error(f"Failed to update config: {str(e)}")
            raise ConfigurationError("Failed to update configuration.")

    @staticmethod
    def _is_instance_active(url: str) -> bool:
        """
        Returns True if the given Opik URL responds to an HTTP GET request.

        Args:
            url (str): The base URL of the instance to check.

        Returns:
            bool: True if the instance responds with HTTP status 200, otherwise False.
        """
        try:
            with httpx.Client(timeout=HEALTH_CHECK_TIMEOUT) as http_client:
                response = http_client.get(url=url + HEALTH_CHECK_URL_POSTFIX)
            return response.status_code == 200
        except httpx.ConnectTimeout:
            return False
        except Exception:
            return False

    def _ask_for_url(self) -> None:
        """
        Prompt the user for an Opik instance URL and check if it is accessible.
        The function retries up to 3 times if the URL is not accessible.

        Raises:
            ConfigurationError: Raised if the URL provided by the user is not accessible after 3 attempts.
        """
        retries = 3
        while retries > 0:
            user_input_opik_url = input("Please enter your Opik instance URL:")
            if self._is_instance_active(user_input_opik_url):
                self.url = user_input_opik_url
                return
            else:
                LOGGER.error(
                    f"Opik is not accessible at {user_input_opik_url}. "
                    f"Please try again, the URL should follow a format similar to {OPIK_BASE_URL_LOCAL}"
                )
                retries -= 1
        raise ConfigurationError(
            "Cannot use the URL provided by the user. Opik instance is not active or not found."
        )

    @property
    def _url_account_details(self) -> str:
        if self.url is None:
            return URL_ACCOUNT_DETAILS_DEFAULT

        if self.url == OPIK_BASE_URL_CLOUD:
            return URL_ACCOUNT_DETAILS_DEFAULT

        return f"{self.url}{URL_ACCOUNT_DETAILS_POSTFIX}"

    @property
    def _url_get_workspace_list(self) -> str:
        if self.url is None:
            return URL_WORKSPACE_GET_LIST_DEFAULT

        if self.url == OPIK_BASE_URL_CLOUD:
            return URL_WORKSPACE_GET_LIST_DEFAULT

        return f"{self.url}{URL_WORKSPACE_GET_LIST_POSTFIX}"


def configure(
    api_key: Optional[str] = None,
    workspace: Optional[str] = None,
    url: Optional[str] = None,
    use_local: bool = False,
    force: bool = False,
) -> None:
    """
    Create a local configuration file for the Python SDK. If a configuration file already exists,
    it will not be overwritten unless the `force` parameter is set to True.

    Args:
        api_key: The API key if using an Opik Cloud.
        workspace: The workspace name if using an Opik Cloud.
        url: The URL of the Opik instance if you are using a local deployment.
        use_local: Whether to use a local deployment.
        force: If true, the configuration file will be recreated and existing settings
               will be overwritten with passed parameters.

    Raises:
        ConfigurationError
    """
    client = OpikConfigurator(
        api_key=api_key,
        workspace=workspace,
        url=url,
        use_local=use_local,
        force=force,
    )
    client.configure()



import abc


class OpikBaseModel(abc.ABC):
    """
    This class serves as an interface to LLMs.

    By the moment it's updated last time, str -> str relation
    for generate and agenerate methods is enough, but it may
    evolve in future if it becomes clear that we need models
    not just consuming or returning strings only.
    """

    def __init__(self, model_name: str):
        self.model_name = model_name

    @abc.abstractmethod
    def generate(self, input: str) -> str: ...

    @abc.abstractmethod
    async def agenerate(self, input: str) -> str: ...



import configparser
import logging
import pathlib
from typing import Any, Dict, Final, List, Literal, Optional, Tuple, Type, Union

import pydantic_settings
from pydantic_settings import BaseSettings, InitSettingsSource
from pydantic_settings.sources import ConfigFileSourceMixin

from . import dict_utils

PathType = Union[
    pathlib.Path,
    str,
    List[Union[pathlib.Path, str]],
    Tuple[Union[pathlib.Path, str], ...],
]

_SESSION_CACHE_DICT: Dict[str, Any] = {}

OPIK_BASE_URL_CLOUD: Final[str] = "https://www.comet.com/opik/api"
OPIK_BASE_URL_LOCAL: Final[str] = "http://localhost:5173/api"

OPIK_PROJECT_DEFAULT_NAME: Final[str] = "Default Project"
OPIK_WORKSPACE_DEFAULT_NAME: Final[str] = "default"

CONFIG_FILE_PATH_DEFAULT: Final[str] = "~/.opik.config"

LOGGER = logging.getLogger(__name__)


class IniConfigSettingsSource(InitSettingsSource, ConfigFileSourceMixin):
    """
    A source class that loads variables from a INI file
    """

    def __init__(
        self,
        settings_cls: Type[BaseSettings],
    ):
        self.ini_data = self._read_files(CONFIG_FILE_PATH_DEFAULT)
        super().__init__(settings_cls, self.ini_data)

    def _read_file(self, file_path: pathlib.Path) -> Dict[str, Any]:
        config = configparser.ConfigParser()
        config.read(file_path)
        config_values = {
            section: dict(config.items(section)) for section in config.sections()
        }

        if "opik" in config_values:
            return config_values["opik"]

        return {}


class OpikConfig(pydantic_settings.BaseSettings):
    """
    Initializes every configuration variable with the first
    found value. The order of sources used:
    1. User passed values
    2. Session config dict (can be populated by calling `update_session_config(...)`)
    3. Environment variables (they must start with "OPIK_" prefix)
    4. Load from file
    5. Default values
    """

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="opik_")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[pydantic_settings.BaseSettings],
        init_settings: pydantic_settings.PydanticBaseSettingsSource,
        env_settings: pydantic_settings.PydanticBaseSettingsSource,
        dotenv_settings: pydantic_settings.PydanticBaseSettingsSource,
        file_secret_settings: pydantic_settings.PydanticBaseSettingsSource,
    ) -> Tuple[pydantic_settings.PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            pydantic_settings.InitSettingsSource(
                pydantic_settings.BaseSettings, _SESSION_CACHE_DICT
            ),
            env_settings,
            IniConfigSettingsSource(settings_cls=cls),
        )

    # Below are Opik configurations

    url_override: str = OPIK_BASE_URL_CLOUD
    """Opik backend base URL"""

    project_name: str = OPIK_PROJECT_DEFAULT_NAME
    """Opik project name"""

    workspace: str = OPIK_WORKSPACE_DEFAULT_NAME
    """Opik workspace"""

    api_key: Optional[str] = None
    """Opik API key. This is not required if you are running against open source Opik installation"""

    default_flush_timeout: Optional[int] = None
    """
    Maximum time to wait when flushing Opik messages queues (in seconds).
    In particular waiting happens when calling:
    * Opik().flush()
    * Opik().end()
    * flush_tracker()
    And when the process is ending.

    If it's not set - there is no timeout.
    """

    background_workers: int = 8
    """
    The amount of background threads that submit data to the backend.
    """

    console_logging_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = (
        "INFO"
    )
    """
    Logging level for console logs.
    """

    file_logging_level: Optional[
        Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    ] = None
    """
    Logging level for file logs. Is not configured - nothing is logged to the file.
    """

    logging_file: str = "opik.log"
    """
    File to write the logs to.
    """

    pytest_experiment_enabled: bool = True
    """
    If enabled, tests decorated with `llm_unit` will log data to Opik experiments
    """

    @property
    def config_file_fullpath(self) -> pathlib.Path:
        return pathlib.Path(CONFIG_FILE_PATH_DEFAULT).expanduser()

    def save_to_file(self) -> None:
        """
        Save configuration to a file

        Raises:
            OSError: If there is an issue writing to the file.
        """
        config_file_content = configparser.ConfigParser()

        config_file_content["opik"] = {
            "url_override": self.url_override,
            "workspace": self.workspace,
        }

        if self.api_key is not None:
            config_file_content["opik"]["api_key"] = self.api_key

        try:
            with open(
                self.config_file_fullpath, mode="w+", encoding="utf-8"
            ) as config_file:
                config_file_content.write(config_file)
            LOGGER.info(f"Configuration saved to file: {self.config_file_fullpath}")
        except OSError as e:
            LOGGER.error(f"Failed to save configuration: {e}")
            raise


def update_session_config(key: str, value: Any) -> None:
    _SESSION_CACHE_DICT[key] = value


def get_from_user_inputs(**user_inputs: Any) -> OpikConfig:
    """
    Instantiates an OpikConfig using provided user inputs.
    """
    cleaned_user_inputs = dict_utils.remove_none_from_dict(user_inputs)

    return OpikConfig(**cleaned_user_inputs)



import logging
import dataclasses
from typing import Any, List, Optional, Dict
from openai.types.chat import chat_completion_chunk

from opik import logging_messages

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class ExtractedStreamContent:
    choices: List[Dict[str, Any]] = dataclasses.field(default_factory=list)
    usage: Optional[Dict[str, Any]] = None


def aggregate(
    items: List[chat_completion_chunk.ChatCompletionChunk],
) -> ExtractedStreamContent:
    extracted_content = ExtractedStreamContent()

    # TODO: check if there are scenarios when stream contains more than one choice
    try:
        content_items = [_get_item_content(item) for item in items]
        choices = [_construct_choice_dict(items[0], content_items)]
        usage = items[-1].usage
        if usage is not None:
            usage = usage.model_dump()

        extracted_content = ExtractedStreamContent(
            choices=choices,
            usage=usage,
        )
    except Exception as exception:
        LOGGER.error(
            logging_messages.FAILED_TO_PARSE_OPENAI_STREAM_CONTENT,
            str(exception),
            exc_info=True,
        )

    return extracted_content


def _get_item_content(item: chat_completion_chunk.ChatCompletionChunk) -> str:
    result: Optional[str] = None
    if len(item.choices) > 0:
        result = item.choices[0].delta.content

    return "" if result is None else result


def _construct_choice_dict(
    first_item: chat_completion_chunk.ChatCompletionChunk, items_content: List[str]
) -> Dict[str, Any]:
    if len(first_item.choices) > 0:
        role = first_item.choices[0].delta.role
    else:
        role = None

    choice_info = {
        "message": {"content": "".join(items_content), "role": role},
    }

    return choice_info



from typing import List, Callable, Dict, Any
import inspect
from opik import exceptions


def raise_if_score_arguments_are_missing(
    score_function: Callable, score_name: str, kwargs: Dict[str, Any]
) -> None:
    signature = inspect.signature(score_function)

    parameters = signature.parameters

    missing_required_arguments: List[str] = []

    for name, param in parameters.items():
        if name == "self":
            continue

        if param.default == inspect.Parameter.empty and param.kind in (
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        ):
            if name not in kwargs:
                missing_required_arguments.append(name)

    if len(missing_required_arguments) > 0:
        raise exceptions.ScoreMethodMissingArguments(
            f"The scoring object {score_name} is missing arguments: {missing_required_arguments}. "
            f"These keys were not present in the dictionary returned by the evaluation task. "
            f"Evaluation task dictionary keys found: {list(kwargs.keys())}."
        )



from typing import Optional, Dict, Any
import pydantic
from .. import constants


class DatasetItem(pydantic.BaseModel):
    """A DatasetItem object representing an item in a dataset."""

    model_config = pydantic.ConfigDict(strict=True)

    input: Dict[str, Any]
    """The input data for the dataset item."""

    expected_output: Optional[Dict[str, Any]] = None
    """The expected output for the dataset item."""

    metadata: Optional[Dict[str, Any]] = None
    """Additional metadata associated with the dataset item."""

    trace_id: Optional[str] = None
    """The ID of the trace associated with this dataset item."""

    span_id: Optional[str] = None
    """The ID of the span associated with this dataset item."""

    id: Optional[str] = None
    """The unique identifier for this dataset item."""

    source: str = constants.DATASET_SOURCE_SDK
    """The source of the dataset item. Defaults to DATASET_SOURCE_SDK."""

    def get_content(self) -> Dict[str, Any]:
        return {
            "input": self.input,
            "expected_output": self.expected_output,
            "metadata": self.metadata,
        }



import inspect

from typing import Callable, Tuple, Any, Dict


def extract_inputs(
    func: Callable, args: Tuple, kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    sig = inspect.signature(func)

    bound_args = sig.bind(*args, **kwargs)  # type: ignore
    bound_args.apply_defaults()

    arg_dict = dict(bound_args.arguments)

    if "self" in arg_dict:
        arg_dict.pop("self")
    elif "cls" in arg_dict:
        arg_dict.pop("cls")

    return arg_dict


def is_generator(obj: Any) -> bool:
    return inspect.isgenerator(obj)


def is_async(func: Callable) -> bool:
    if inspect.iscoroutinefunction(func):
        return True

    # Usual iscoroutinefunction might not work correctly if
    # async function is wrapped.
    if hasattr(func, "__wrapped__") and inspect.iscoroutinefunction(func.__wrapped__):
        return True

    return False



import functools
import atexit
import datetime
import logging

from typing import Optional, Any, Dict, List, Mapping
from ..types import SpanType, UsageDict, FeedbackScoreDict
from . import (
    span,
    trace,
    dataset,
    experiment,
    helpers,
    constants,
    validation_helpers,
)
from ..message_processing import streamer_constructors, messages, jsonable_encoder
from ..rest_api import client as rest_api_client
from ..rest_api.types import dataset_public, trace_public, span_public
from .. import datetime_helpers, config, httpx_client


LOGGER = logging.getLogger(__name__)


class Opik:
    def __init__(
        self,
        project_name: Optional[str] = None,
        workspace: Optional[str] = None,
        host: Optional[str] = None,
        _use_batching: bool = False,
    ) -> None:
        """
        Initialize an Opik object that can be used to log traces and spans manually to Opik server.

        Args:
            project_name: The name of the project. If not provided, traces and spans will be logged to the `Default Project`.
            workspace: The name of the workspace. If not provided, `default` will be used.
            host: The host URL for the Opik server. If not provided, it will default to `https://www.comet.com/opik/api`.
            _use_batching: intended for internal usage in specific conditions only.
                Enabling it is unsafe and can lead to data loss.
        Returns:
            None
        """
        config_ = config.get_from_user_inputs(
            project_name=project_name, workspace=workspace, url_override=host
        )
        self._workspace: str = config_.workspace
        self._project_name: str = config_.project_name
        self._flush_timeout: Optional[int] = config_.default_flush_timeout

        self._initialize_streamer(
            base_url=config_.url_override,
            workers=config_.background_workers,
            api_key=config_.api_key,
            use_batching=_use_batching,
        )
        atexit.register(self.end, timeout=self._flush_timeout)

    def _initialize_streamer(
        self,
        base_url: str,
        workers: int,
        api_key: Optional[str],
        use_batching: bool,
    ) -> None:
        httpx_client_ = httpx_client.get(workspace=self._workspace, api_key=api_key)
        self._rest_client = rest_api_client.OpikApi(
            base_url=base_url,
            httpx_client=httpx_client_,
        )

        self._streamer = streamer_constructors.construct_online_streamer(
            n_consumers=workers,
            rest_client=self._rest_client,
            use_batching=use_batching,
        )

    def trace(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        feedback_scores: Optional[List[FeedbackScoreDict]] = None,
    ) -> trace.Trace:
        """
        Create and log a new trace.

        Args:
            id: The unique identifier for the trace, if not provided a new ID will be generated. Must be a valid [UUIDv8](https://uuid.ramsey.dev/en/stable/rfc4122/version8.html) ID.
            name: The name of the trace.
            start_time: The start time of the trace. If not provided, the current local time will be used.
            end_time: The end time of the trace.
            input: The input data for the trace. This can be any valid JSON serializable object.
            output: The output data for the trace. This can be any valid JSON serializable object.
            metadata: Additional metadata for the trace. This can be any valid JSON serializable object.
            tags: Tags associated with the trace.
            feedback_scores: The list of feedback score dicts assosiated with the trace. Dicts don't required to have an `id` value.

        Returns:
            trace.Trace: The created trace object.
        """
        id = id if id is not None else helpers.generate_id()
        start_time = (
            start_time if start_time is not None else datetime_helpers.local_timestamp()
        )
        create_trace_message = messages.CreateTraceMessage(
            trace_id=id,
            project_name=self._project_name,
            name=name,
            start_time=start_time,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            tags=tags,
        )
        self._streamer.put(create_trace_message)

        if feedback_scores is not None:
            for feedback_score in feedback_scores:
                feedback_score["id"] = id

            self.log_traces_feedback_scores(feedback_scores)

        return trace.Trace(
            id=id,
            message_streamer=self._streamer,
            project_name=self._project_name,
        )

    def span(
        self,
        trace_id: Optional[str] = None,
        id: Optional[str] = None,
        parent_span_id: Optional[str] = None,
        name: Optional[str] = None,
        type: SpanType = "general",
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        usage: Optional[UsageDict] = None,
        feedback_scores: Optional[List[FeedbackScoreDict]] = None,
    ) -> span.Span:
        """
        Create and log a new span.

        Args:
            trace_id: The unique identifier for the trace. If not provided, a new ID will be generated. Must be a valid [UUIDv8](https://uuid.ramsey.dev/en/stable/rfc4122/version8.html) ID.
            id: The unique identifier for the span. If not provided, a new ID will be generated. Must be a valid [UUIDv8](https://uuid.ramsey.dev/en/stable/rfc4122/version8.html) ID.
            parent_span_id: The unique identifier for the parent span.
            name: The name of the span.
            type: The type of the span. Default is "general".
            start_time: The start time of the span. If not provided, the current local time will be used.
            end_time: The end time of the span.
            metadata: Additional metadata for the span. This can be any valid JSON serializable object.
            input: The input data for the span. This can be any valid JSON serializable object.
            output: The output data for the span. This can be any valid JSON serializable object.
            tags: Tags associated with the span.
            usage: Usage data for the span.
            feedback_scores: The list of feedback score dicts assosiated with the span. Dicts don't required to have an `id` value.

        Returns:
            span.Span: The created span object.
        """
        id = id if id is not None else helpers.generate_id()
        start_time = (
            start_time if start_time is not None else datetime_helpers.local_timestamp()
        )

        parsed_usage = validation_helpers.validate_and_parse_usage(usage, LOGGER)
        if parsed_usage.full_usage is not None:
            metadata = (
                {"usage": parsed_usage.full_usage}
                if metadata is None
                else {"usage": parsed_usage.full_usage, **metadata}
            )

        if trace_id is None:
            trace_id = helpers.generate_id()
            # TODO: decide what needs to be passed to CreateTraceMessage.
            # This version is likely not final.
            create_trace_message = messages.CreateTraceMessage(
                trace_id=trace_id,
                project_name=self._project_name,
                name=name,
                start_time=start_time,
                end_time=end_time,
                input=input,
                output=output,
                metadata=metadata,
                tags=tags,
            )
            self._streamer.put(create_trace_message)

        create_span_message = messages.CreateSpanMessage(
            span_id=id,
            trace_id=trace_id,
            project_name=self._project_name,
            parent_span_id=parent_span_id,
            name=name,
            type=type,
            start_time=start_time,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            tags=tags,
            usage=parsed_usage.supported_usage,
        )
        self._streamer.put(create_span_message)

        if feedback_scores is not None:
            for feedback_score in feedback_scores:
                feedback_score["id"] = id

            self.log_spans_feedback_scores(feedback_scores)

        return span.Span(
            id=id,
            parent_span_id=parent_span_id,
            trace_id=trace_id,
            project_name=self._project_name,
            message_streamer=self._streamer,
        )

    def log_spans_feedback_scores(self, scores: List[FeedbackScoreDict]) -> None:
        """
        Log feedback scores for spans.

        Args:
            scores (List[FeedbackScoreDict]): A list of feedback score dictionaries.
                Specifying a span id via `id` key for each score is mandatory.

        Returns:
            None
        """
        valid_scores = [
            score
            for score in scores
            if validation_helpers.validate_feedback_score(score, LOGGER) is not None
        ]

        if len(valid_scores) == 0:
            return None

        score_messages = [
            messages.FeedbackScoreMessage(
                source=constants.FEEDBACK_SCORE_SOURCE_SDK,
                project_name=self._project_name,
                **score_dict,
            )
            for score_dict in valid_scores
        ]

        for batch in helpers.list_to_batches(
            score_messages, batch_size=constants.FEEDBACK_SCORES_MAX_BATCH_SIZE
        ):
            add_span_feedback_scores_batch_message = (
                messages.AddSpanFeedbackScoresBatchMessage(batch=batch)
            )

            self._streamer.put(add_span_feedback_scores_batch_message)

    def log_traces_feedback_scores(self, scores: List[FeedbackScoreDict]) -> None:
        """
        Log feedback scores for traces.

        Args:
            scores (List[FeedbackScoreDict]): A list of feedback score dictionaries.
                Specifying a trace id via `id` key for each score is mandatory.

        Returns:
            None
        """
        valid_scores = [
            score
            for score in scores
            if validation_helpers.validate_feedback_score(score, LOGGER) is not None
        ]

        if len(valid_scores) == 0:
            return None

        score_messages = [
            messages.FeedbackScoreMessage(
                source=constants.FEEDBACK_SCORE_SOURCE_SDK,
                project_name=self._project_name,
                **score_dict,
            )
            for score_dict in valid_scores
        ]
        for batch in helpers.list_to_batches(
            score_messages, batch_size=constants.FEEDBACK_SCORES_MAX_BATCH_SIZE
        ):
            add_span_feedback_scores_batch_message = (
                messages.AddTraceFeedbackScoresBatchMessage(batch=batch)
            )

            self._streamer.put(add_span_feedback_scores_batch_message)

    def get_dataset(self, name: str) -> dataset.Dataset:
        """
        Get dataset by name

        Args:
            name: The name of the dataset

        Returns:
            dataset.Dataset: dataset object associated with the name passed.
        """
        dataset_fern: dataset_public.DatasetPublic = (
            self._rest_client.datasets.get_dataset_by_identifier(dataset_name=name)
        )

        dataset_ = dataset.Dataset(
            name=name,
            description=dataset_fern.description,
            rest_client=self._rest_client,
        )

        return dataset_

    def delete_dataset(self, name: str) -> None:
        """
        Delete dataset by name

        Args:
            name: The name of the dataset
        """
        self._rest_client.datasets.delete_dataset_by_name(dataset_name=name)

    def create_dataset(
        self, name: str, description: Optional[str] = None
    ) -> dataset.Dataset:
        """
        Create a new dataset.

        Args:
            name: The name of the dataset.
            description: An optional description of the dataset.

        Returns:
            dataset.Dataset: The created dataset object.
        """
        self._rest_client.datasets.create_dataset(name=name, description=description)

        result = dataset.Dataset(
            name=name,
            description=description,
            rest_client=self._rest_client,
        )

        return result

    def create_experiment(
        self,
        name: str,
        dataset_name: str,
        experiment_config: Optional[Dict[str, Any]] = None,
    ) -> experiment.Experiment:
        id = helpers.generate_id()

        if isinstance(experiment_config, Mapping):
            metadata = jsonable_encoder.jsonable_encoder(experiment_config)
        elif experiment_config is not None:
            LOGGER.error(
                "Experiment config must be dictionary, but %s was provided. Config will not be logged.",
                experiment_config,
            )
            metadata = None
        else:
            metadata = None

        self._rest_client.experiments.create_experiment(
            name=name,
            dataset_name=dataset_name,
            id=id,
            metadata=metadata,
        )

        experiment_ = experiment.Experiment(
            id=id,
            name=name,
            dataset_name=dataset_name,
            rest_client=self._rest_client,
        )

        return experiment_

    def end(self, timeout: Optional[int] = None) -> None:
        """
        End the Opik session and submit all pending messages.

        Args:
            timeout (Optional[int]): The timeout for closing the streamer. Once the timeout is reached, the streamer will be closed regardless of whether all messages have been sent. If no timeout is set, the default value from the Opik configuration will be used.

        Returns:
            None
        """
        timeout = timeout if timeout is not None else self._flush_timeout
        self._streamer.close(timeout)

    def flush(self, timeout: Optional[int] = None) -> None:
        """
        Flush the streamer to ensure all messages are sent.

        Args:
            timeout (Optional[int]): The timeout for flushing the streamer. Once the timeout is reached, the flush method will return regardless of whether all messages have been sent.

        Returns:
            None
        """
        timeout = timeout if timeout is not None else self._flush_timeout
        self._streamer.flush(timeout)

    def get_trace_content(self, id: str) -> trace_public.TracePublic:
        """
        Args:
            id (str): trace id
        Returns:
            trace_public.TracePublic: pydantic model object with all the data associated with the trace found.
            Raises an error if trace was not found.
        """
        return self._rest_client.traces.get_trace_by_id(id)

    def get_span_content(self, id: str) -> span_public.SpanPublic:
        """
        Args:
            id (str): span id
        Returns:
            span_public.SpanPublic: pydantic model object with all the data associated with the span found.
            Raises an error if span was not found.
        """
        return self._rest_client.spans.get_span_by_id(id)


@functools.lru_cache()
def get_client_cached() -> Opik:
    client = Opik(_use_batching=True)

    return client



import json
import logging
from typing import Union, Optional, List, Any

from opik.evaluation.models import base_model, models_factory
from opik.evaluation.metrics import score_result, base_metric
from opik import logging_messages

from . import template
from ... import exceptions

LOGGER = logging.getLogger(__name__)


class Hallucination(base_metric.BaseMetric):
    """
    A metric that evaluates whether an LLM's output contains hallucinations based on given input and context.

    This metric uses another LLM to judge if the output is factual or contains hallucinations.
    It returns a score of 1.0 if hallucination is detected, and 0.0 otherwise.

    Args:
        model: The LLM to use for evaluation. Can be a string (model name) or a CometBaseModel instance.
        name: The name of the metric.
        few_shot_examples: A list of few-shot examples to use for hallucination detection.  If None, default examples will be used.

    Example:
        >>> from comet_llm_eval.evaluation.metrics import Hallucination
        >>> hallucination_metric = Hallucination()
        >>> result = hallucination_metric.score(
        ...     input="What is the capital of France?",
        ...     output="The capital of France is London.",
        ...     context=["The capital of France is Paris."]
        ... )
        >>> print(result.value)
        1.0
        >>> print(result.reason)
        The answer provided states that the capital of France is London, which contradicts the fact stated in the context that the capital of France is Paris.
    """

    def __init__(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]] = None,
        name: str = "hallucination_metric",
        few_shot_examples: Optional[List[template.FewShotExampleHallucination]] = None,
    ):
        super().__init__(name=name)
        self._init_model(model)
        self.few_shot_examples = [] if few_shot_examples is None else few_shot_examples

    def _init_model(
        self, model: Optional[Union[str, base_model.OpikBaseModel]]
    ) -> None:
        if isinstance(model, base_model.OpikBaseModel):
            self._model = model
        else:
            self._model = models_factory.get(model_name=model)

    def score(
        self, input: str, output: str, context: List[str], **ignored_kwargs: Any
    ) -> score_result.ScoreResult:
        """
        Calculate the hallucination score for the given input, output, and context.

        Args:
            input: The original input/question.
            output: The LLM's output to evaluate.
            context: A list of context strings.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with a value of 1.0 if hallucination
                is detected, 0.0 otherwise, along with the reason for the verdict.
        """
        llm_query = template.generate_query(
            input=input,
            output=output,
            context=context,
            few_shot_examples=self.few_shot_examples,
        )
        model_output = self._model.generate(input=llm_query)

        return self._parse_model_output(model_output)

    async def ascore(
        self, input: str, output: str, context: List[str], **ignored_kwargs: Any
    ) -> score_result.ScoreResult:
        """
        Asynchronously calculate the hallucination score for the given input, output, and context.

        Args:
            input: The original input/question.
            output: The LLM's output to evaluate.
            context: A list of context strings.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with a value of 1.0 if hallucination
                is detected, 0.0 otherwise, along with the reason for the verdict.
        """
        llm_query = template.generate_query(
            input=input,
            output=output,
            context=context,
            few_shot_examples=self.few_shot_examples,
        )
        model_output = await self._model.agenerate(input=llm_query)

        return self._parse_model_output(model_output)

    def _parse_model_output(self, content: str) -> score_result.ScoreResult:
        try:
            dict_content = json.loads(content)
            verdict: str = dict_content[template.VERDICT_KEY]
            score = 1.0 if verdict.lower() == template.HALLUCINATION_VERDICT else 0.0
            return score_result.ScoreResult(
                name=self.name,
                value=score,
                reason=str(dict_content[template.REASON_KEY]),
            )
        except Exception:
            raise exceptions.MetricComputationError(
                logging_messages.HALLUCINATION_DETECTION_FAILED
            )



import queue
from typing import Type, Dict

from .. import messages

from . import base_batcher
from . import create_span_message_batcher
from . import batch_manager

CREATE_SPANS_MESSAGE_BATCHER_FLUSH_INTERVAL_SECONDS = 1.0
CREATE_SPANS_MESSAGE_BATCHER_MAX_BATCH_SIZE = 1000


def create_batch_manager(message_queue: queue.Queue) -> batch_manager.BatchManager:
    create_span_message_batcher_ = create_span_message_batcher.CreateSpanMessageBatcher(
        flush_interval_seconds=CREATE_SPANS_MESSAGE_BATCHER_FLUSH_INTERVAL_SECONDS,
        max_batch_size=CREATE_SPANS_MESSAGE_BATCHER_MAX_BATCH_SIZE,
        flush_callback=message_queue.put,
    )

    MESSAGE_TO_BATCHER_MAPPING: Dict[
        Type[messages.BaseMessage], base_batcher.BaseBatcher
    ] = {messages.CreateSpanMessage: create_span_message_batcher_}

    batch_manager_ = batch_manager.BatchManager(
        message_to_batcher_mapping=MESSAGE_TO_BATCHER_MAPPING
    )

    return batch_manager_



import logging
import json
from typing import Optional, Any, List, Dict, Union, Sequence

from opik.rest_api import client as rest_api_client
from opik.rest_api.types import dataset_item as rest_dataset_item
from opik import exceptions

from .. import helpers, constants
from . import dataset_item, converters

import pandas

LOGGER = logging.getLogger(__name__)


class Dataset:
    def __init__(
        self,
        name: str,
        description: Optional[str],
        rest_client: rest_api_client.OpikApi,
    ) -> None:
        """
        A Dataset object. This object should not be created directly, instead use :meth:`opik.Opik.create_dataset` or :meth:`opik.Opik.get_dataset`.
        """
        self._name = name
        self._description = description
        self._rest_client = rest_client

    @property
    def name(self) -> str:
        """The name of the dataset."""
        return self._name

    @property
    def description(self) -> Optional[str]:
        """The description of the dataset."""
        return self._description

    def insert(
        self, items: Sequence[Union[dataset_item.DatasetItem, Dict[str, Any]]]
    ) -> None:
        """
        Insert new items into the dataset.

        Args:
            items: List of DatasetItem objects or dicts (which will be converted to DatasetItem objects)
                to add to the dataset.
        """
        items: List[dataset_item.DatasetItem] = [  # type: ignore
            (dataset_item.DatasetItem(**item) if isinstance(item, dict) else item)
            for item in items
        ]

        rest_items = [
            rest_dataset_item.DatasetItem(
                id=item.id if item.id is not None else helpers.generate_id(),  # type: ignore
                input=item.input,  # type: ignore
                expected_output=item.expected_output,  # type: ignore
                metadata=item.metadata,  # type: ignore
                trace_id=item.trace_id,  # type: ignore
                span_id=item.span_id,  # type: ignore
                source=item.source,  # type: ignore
            )
            for item in items
        ]

        batches = helpers.list_to_batches(
            rest_items, batch_size=constants.DATASET_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Sending dataset items batch: %s", batch)
            self._rest_client.datasets.create_or_update_dataset_items(
                dataset_name=self._name, items=batch
            )

    def update(self, items: List[dataset_item.DatasetItem]) -> None:
        """
        Update existing items in the dataset.

        Args:
            items: List of DatasetItem objects to update in the dataset. You need to provide the full item object as it will override what has been supplied previously.

        Raises:
            DatasetItemUpdateOperationRequiresItemId: If any item in the list is missing an id.
        """
        for item in items:
            if item.id is None:  # TODO: implement proper exception
                raise exceptions.DatasetItemUpdateOperationRequiresItemId(
                    "Missing id for dataset item to update: %s", item
                )

        self.insert(items)

    def delete(self, items_ids: List[str]) -> None:
        """
        Delete items from the dataset.

        Args:
            items_ids: List of item ids to delete.
        """
        batches = helpers.list_to_batches(
            items_ids, batch_size=constants.DATASET_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Deleting dataset items batch: %s", batch)
            self._rest_client.datasets.delete_dataset_items(item_ids=batch)

    def clear(self) -> None:
        """
        Delete all items from the given dataset.
        """
        all_items = self.get_all_items()
        item_ids = [item.id for item in all_items if item.id is not None]
        self.delete(item_ids)

    def to_pandas(self) -> pandas.DataFrame:
        """
        Convert the dataset to a pandas DataFrame.

        Returns:
            A pandas DataFrame containing all items in the dataset.
        """
        dataset_items = self.get_all_items()

        return converters.to_pandas(dataset_items, keys_mapping={})

    def to_json(self) -> str:
        """
        Convert the dataset to a JSON string.

        Returns:
            A JSON string representation of all items in the dataset.
        """
        dataset_items = self.get_all_items()

        return converters.to_json(dataset_items, keys_mapping={})

    def get_all_items(self) -> List[dataset_item.DatasetItem]:
        """
        Retrieve all items from the dataset.

        Returns:
            A list of DatasetItem objects representing all items in the dataset.
        """
        results: List[dataset_item.DatasetItem] = []

        while True:
            stream = self._rest_client.datasets.stream_dataset_items(
                dataset_name=self._name,
                last_retrieved_id=results[-1].id if len(results) > 0 else None,
            )

            item_bytes = b"".join(stream)
            stream_results: List[dataset_item.DatasetItem] = []
            for line in item_bytes.split(b"\n"):
                if len(line) == 0:
                    continue

                item_content: Dict[str, Any] = json.loads(line.decode("utf-8").strip())

                item = dataset_item.DatasetItem(
                    id=item_content.get("id"),  # type: ignore
                    input=item_content.get("input"),  # type: ignore
                    expected_output=item_content.get("expected_output"),  # type: ignore
                    metadata=item_content.get("metadata"),  # type: ignore
                    trace_id=item_content.get("trace_id"),  # type: ignore
                    span_id=item_content.get("span_id"),  # type: ignore
                    source=item_content.get("source"),  # type: ignore
                )

                stream_results.append(item)

            if len(stream_results) == 0:
                break

            results.extend(stream_results)

        return results

    def insert_from_json(
        self,
        json_array: str,
        keys_mapping: Optional[Dict[str, str]] = None,
        ignore_keys: Optional[List[str]] = None,
    ) -> None:
        """
        Args:
            json_array: json string of format: "[{...}, {...}, {...}]" where every dictionary
                is to be transformed into dataset item
            keys_mapping: dictionary that maps json keys to item fields names
                Example: {'Expected output': 'expected_output'}
            ignore_keys: if your json dicts contain keys that are not needed for DatasetItem
                construction - pass them as ignore_keys argument
        """
        keys_mapping = {} if keys_mapping is None else keys_mapping
        ignore_keys = [] if ignore_keys is None else ignore_keys

        new_items = converters.from_json(
            json_array, keys_mapping=keys_mapping, ignore_keys=ignore_keys
        )

        self.insert(new_items)

    def read_jsonl_from_file(
        self,
        file_path: str,
        keys_mapping: Optional[Dict[str, str]] = None,
        ignore_keys: Optional[List[str]] = None,
    ) -> None:
        """
        Read JSONL from a file and insert it into the dataset.

        Args:
            file_path: Path to the JSONL file
            keys_mapping: dictionary that maps json keys to item fields names
                Example: {'Expected output': 'expected_output'}
            ignore_keys: if your json dicts contain keys that are not needed for DatasetItem
                construction - pass them as ignore_keys argument
        """
        keys_mapping = {} if keys_mapping is None else keys_mapping
        ignore_keys = [] if ignore_keys is None else ignore_keys
        new_items = converters.from_jsonl_file(file_path, keys_mapping, ignore_keys)
        self.insert(new_items)

    def insert_from_pandas(
        self,
        dataframe: pandas.DataFrame,
        keys_mapping: Optional[Dict[str, str]] = None,
        ignore_keys: Optional[List[str]] = None,
    ) -> None:
        """
        Args:
            dataframe: pandas dataframe
            keys_mapping: Dictionary that maps dataframe column names to dataset item field names.
                Example: {'Expected output': 'expected_output'}
            ignore_keys: if your dataframe contains columns that are not needed for DatasetItem
                construction - pass them as ignore_keys argument
        """
        keys_mapping = {} if keys_mapping is None else keys_mapping
        ignore_keys = [] if ignore_keys is None else ignore_keys

        new_items = converters.from_pandas(dataframe, keys_mapping, ignore_keys)

        self.insert(new_items)



from typing import Any

import langchain_openai

from . import base_model


class LangchainChatModel(base_model.OpikBaseModel):
    def __init__(
        self, model_name: str = "gpt-3.5-turbo", **chatopenai_kwargs: Any
    ) -> None:
        super().__init__(model_name=model_name)
        self._engine = langchain_openai.ChatOpenAI(name=model_name, **chatopenai_kwargs)

    def generate(self, input: str) -> str:
        response = self._engine.invoke(input=input)

        return str(response.content)

    async def agenerate(self, input: str) -> str:
        response = await self._engine.ainvoke(input=input)

        return str(response.content)



from typing import Union

import openai

from . import openai_decorator, chunks_aggregator


def track_openai(
    openai_client: Union[openai.OpenAI, openai.AsyncOpenAI],
) -> Union[openai.OpenAI, openai.AsyncOpenAI]:
    """Adds Opik tracking to an OpenAI client.

    Tracks calls to `openai_client.chat.completions.create()`, it includes support for streaming model.
    Can be used within other Opik-tracked functions.

    Args:
        openai_client: An instance of OpenAI or AsyncOpenAI client.

    Returns:
        The modified OpenAI client with Opik tracking enabled.
    """
    decorator = openai_decorator.OpenaiTrackDecorator()
    wrapper = decorator.track(
        type="llm",
        name="chat_completion_create",
        generations_aggregator=chunks_aggregator.aggregate,
    )
    openai_client.chat.completions.create = wrapper(
        openai_client.chat.completions.create
    )

    return openai_client



import pydantic


class TaskOutput(pydantic.BaseModel):
    # Keys that are already used by our metrics.
    #
    # input
    # output
    # expected
    # context
    # metadata

    # Model config allows to provide custom fields.
    # It might be especially relevant for custom metrics.
    model_config = pydantic.ConfigDict(extra="allow", strict=False)



class OpikException(Exception):
    pass


class DatasetItemUpdateOperationRequiresItemId(OpikException):
    pass


class ContextExtractorNotSet(OpikException):
    pass


class ConfigurationError(OpikException):
    pass


class ScoreMethodMissingArguments(OpikException):
    pass



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from ..core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1
from .feedback_score_average_public import FeedbackScoreAveragePublic
from .json_node_public import JsonNodePublic


class ExperimentPublic(pydantic_v1.BaseModel):
    id: typing.Optional[str] = None
    dataset_name: str
    dataset_id: typing.Optional[str] = None
    name: str
    metadata: typing.Optional[JsonNodePublic] = None
    feedback_scores: typing.Optional[typing.List[FeedbackScoreAveragePublic]] = None
    trace_count: typing.Optional[int] = None
    created_at: typing.Optional[dt.datetime] = None
    last_updated_at: typing.Optional[dt.datetime] = None
    created_by: typing.Optional[str] = None
    last_updated_by: typing.Optional[str] = None

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        kwargs_with_defaults_exclude_none: typing.Any = {
            "by_alias": True,
            "exclude_none": True,
            **kwargs,
        }

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset),
            super().dict(**kwargs_with_defaults_exclude_none),
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic

IS_PYDANTIC_V2 = pydantic.VERSION.startswith("2.")

if IS_PYDANTIC_V2:
    import pydantic.v1 as pydantic_v1  # type: ignore  # nopycln: import
else:
    import pydantic as pydantic_v1  # type: ignore  # nopycln: import


def deep_union_pydantic_dicts(
    source: typing.Dict[str, typing.Any], destination: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            deep_union_pydantic_dicts(value, node)
        else:
            destination[key] = value

    return destination


__all__ = ["pydantic_v1"]



from typing import List, TypedDict

VERDICT_KEY = "context_precision_score"
REASON_KEY = "reason"


class FewShotExampleContextPrecision(TypedDict):
    title: str
    input: str
    expected_output: str
    context: str
    output: str
    context_precision_score: float
    reason: str


FEW_SHOT_EXAMPLES: List[FewShotExampleContextPrecision] = [
    {
        "title": "Low ContextPrecision Score",
        "input": "What is the capital of France?",
        "expected_output": "The capital of France is Paris.",
        "context": "The user is asking about the capital city of a European country.",
        "output": "The capital of Italy is Rome.",
        "context_precision_score": 0.0,
        "reason": "The answer provided by the LLM is completely inaccurate as it refers to the wrong country and does not address the user's question about France.",
    },
    {
        "title": "Medium ContextPrecision Score",
        "input": "What is the capital of France?",
        "expected_output": "The capital of France is Paris.",
        "context": "The user is asking for the capital city of France.",
        "output": "Paris is a city in Europe.",
        "context_precision_score": 0.4,
        "reason": "While the answer mentions Paris, it fails to clearly identify it as the capital of France, thus providing only partial accuracy.",
    },
    {
        "title": "High ContextPrecision Score",
        "input": "What is the capital of France?",
        "expected_output": "The capital of France is Paris.",
        "context": "The user is asking about the capital city of France.",
        "output": "The capital of France is Paris.",
        "context_precision_score": 1.0,
        "reason": "The answer perfectly matches the expected response, fully aligning with the context and providing the correct information.",
    },
]


def generate_query(
    input: str,
    output: str,
    expected_output: str,
    context: List[str],
    few_shot_examples: List[FewShotExampleContextPrecision] = FEW_SHOT_EXAMPLES,
) -> str:
    examples_str = "\n\n".join(
        [
            f"#### Example {i+1}: {example['title']}\n\n"
            f"- **Input:** \"{example['input']}\"\n"
            f"- **Output:** \"{example['output']}\"\n"
            f"- **Expected Output:** \"{example['expected_output']}\"\n"
            f"- **Context:** \"{example['context']}\"\n"
            f"- **Result:**\n"
            f"  ```json\n"
            f"  {{\n"
            f"    \"{VERDICT_KEY}\": {example['context_precision_score']},\n"
            f"    \"{REASON_KEY}\": \"{example['reason']}\"\n"
            f"  }}\n"
            f"  ```"
            for i, example in enumerate(few_shot_examples)
        ]
    )

    return f"""YOU ARE AN EXPERT EVALUATOR SPECIALIZED IN ASSESSING THE "CONTEXT PRECISION" METRIC FOR LLM GENERATED OUTPUTS.
YOUR TASK IS TO EVALUATE HOW PRECISELY A GIVEN ANSWER FROM AN LLM FITS THE EXPECTED ANSWER, GIVEN THE CONTEXT AND USER INPUT.

###INSTRUCTIONS###

1. **EVALUATE THE CONTEXT PRECISION:**
    - **ANALYZE** the provided user input, expected answer, answer from another LLM, and the context.
    - **COMPARE** the answer from the other LLM with the expected answer, focusing on how well it aligns in terms of context, relevance, and accuracy.
    - **ASSIGN A SCORE** from 0.0 to 1.0 based on the following scale:

###SCALE FOR CONTEXT PRECISION METRIC (0.0 - 1.0)###

- **0.0:** COMPLETELY INACCURATE – The LLM's answer is entirely off-topic, irrelevant, or incorrect based on the context and expected answer.
- **0.2:** MOSTLY INACCURATE – The answer contains significant errors, misunderstanding of the context, or is largely irrelevant.
- **0.4:** PARTIALLY ACCURATE – Some correct elements are present, but the answer is incomplete or partially misaligned with the context and expected answer.
- **0.6:** MOSTLY ACCURATE – The answer is generally correct and relevant but may contain minor errors or lack complete precision in aligning with the expected answer.
- **0.8:** HIGHLY ACCURATE – The answer is very close to the expected answer, with only minor discrepancies that do not significantly impact the overall correctness.
- **1.0:** PERFECTLY ACCURATE – The LLM's answer matches the expected answer precisely, with full adherence to the context and no errors.

2. **PROVIDE A REASON FOR THE SCORE:**
    - **JUSTIFY** why the specific score was given, considering the alignment with context, accuracy, relevance, and completeness.

3. **RETURN THE RESULT IN A JSON FORMAT** as follows:
    - `"{VERDICT_KEY}"`: The score between 0.0 and 1.0.
    - `"{REASON_KEY}"`: A detailed explanation of why the score was assigned.

###WHAT NOT TO DO###
- **DO NOT** assign a high score to answers that are off-topic or irrelevant, even if they contain some correct information.
- **DO NOT** give a low score to an answer that is nearly correct but has minor errors or omissions; instead, accurately reflect its alignment with the context.
- **DO NOT** omit the justification for the score; every score must be accompanied by a clear, reasoned explanation.
- **DO NOT** disregard the importance of context when evaluating the precision of the answer.
- **DO NOT** assign scores outside the 0.0 to 1.0 range.
- **DO NOT** return any output format other than JSON.

###FEW-SHOT EXAMPLES###

{examples_str}

NOW, EVALUATE THE PROVIDED INPUTS AND CONTEXT TO DETERMINE THE CONTEXT PRECISION SCORE.

###INPUTS:###
***
Input:
{input}

Output:
{output}

Expected Output:
{expected_output}

Context:
{context}
***
    """



import logging
from typing import List

from opik.rest_api import client as rest_api_client
from opik.rest_api.types import experiment_item as rest_experiment_item
from . import experiment_item
from .. import helpers, constants


LOGGER = logging.getLogger(__name__)


class Experiment:
    def __init__(
        self,
        id: str,
        name: str,
        dataset_name: str,
        rest_client: rest_api_client.OpikApi,
    ) -> None:
        self.id = id
        self.name = name
        self.dataset_name = dataset_name
        self._rest_client = rest_client

    def insert(self, experiment_items: List[experiment_item.ExperimentItem]) -> None:
        rest_experiment_items = [
            rest_experiment_item.ExperimentItem(
                id=item.id if item.id is not None else helpers.generate_id(),
                experiment_id=self.id,
                dataset_item_id=item.dataset_item_id,
                trace_id=item.trace_id,
            )
            for item in experiment_items
        ]

        batches = helpers.list_to_batches(
            rest_experiment_items, batch_size=constants.EXPERIMENT_ITEMS_MAX_BATCH_SIZE
        )

        for batch in batches:
            LOGGER.debug("Sending experiment items batch: %s", batch)
            self._rest_client.experiments.create_experiment_items(
                experiment_items=batch,
            )



import pydantic
import dataclasses

from typing import Any, Dict, Optional
from ..types import UsageDict
from . import validator, result


class PydanticWrapper(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid")
    usage: UsageDict


@dataclasses.dataclass
class ParsedUsage:
    full_usage: Optional[Dict[str, Any]] = None
    supported_usage: Optional[UsageDict] = None


EXPECTED_TYPES = "{'completion_tokens': int, 'prompt_tokens': int, 'total_tokens': int}"


class UsageValidator(validator.Validator):
    """
    Validator for span token usage
    """

    def __init__(self, usage: Any):
        self.usage = usage

        self.parsed_usage = ParsedUsage()

    def validate(self) -> result.ValidationResult:
        try:
            if isinstance(self.usage, dict):
                filtered_usage = _keep_supported_keys(self.usage)
                PydanticWrapper(usage=filtered_usage)
                supported_usage = UsageDict(**filtered_usage)  # type: ignore
                self.parsed_usage = ParsedUsage(
                    full_usage=self.usage, supported_usage=supported_usage
                )
            else:
                # we already know that usage is invalid but want pydantic to trigger validation error
                PydanticWrapper(usage=self.usage)

            self.validation_result = result.ValidationResult(failed=False)
        except pydantic.ValidationError as exception:
            failure_reasons = []
            for e in exception.errors():
                component_name: str = ".".join(e["loc"])
                component_value: str = e["input"]
                msg: str = (
                    f"{component_name} is invalid or missing.\n"
                    f"Expected keys to have in a dictionary: {EXPECTED_TYPES}.\n"
                    f"Value {repr(component_value)} of type {type(component_value)} was passed."
                )
                failure_reasons.append(msg)
            self.validation_result = result.ValidationResult(
                failed=True, failure_reasons=failure_reasons
            )

        return self.validation_result

    def failure_reason_message(self) -> str:
        assert (
            len(self.validation_result.failure_reasons) > 0
        ), "validate() must be called before accessing failure reason message"
        return self.validation_result.failure_reasons[0]


def _keep_supported_keys(usage: Dict[str, Any]) -> Dict[str, Any]:
    supported_keys = UsageDict.__annotations__.keys()
    filtered_usage = {}

    for key in supported_keys:
        if key in usage:
            filtered_usage[key] = usage[key]

    return filtered_usage



import time
from typing import List, Dict, Any, Optional

from .types import LLMTask
from .metrics import base_metric
from ..api_objects.dataset import dataset
from ..api_objects.experiment import experiment_item
from ..api_objects import opik_client

from . import tasks_scorer, scores_logger, report, evaluation_result


def evaluate(
    dataset: dataset.Dataset,
    task: LLMTask,
    scoring_metrics: List[base_metric.BaseMetric],
    experiment_name: str,
    experiment_config: Optional[Dict[str, Any]] = None,
    verbose: int = 1,
    task_threads: int = 16,
) -> evaluation_result.EvaluationResult:
    """
    Performs task evaluation on a given dataset.

    Args:
        dataset: An Opik dataset instance

        task: A callable object that takes DatasetItem as input and returns
            dictionary which will later be used for scoring

        experiment_name: The name of the experiment associated with evaluation run

        experiment_config: The dictionary with parameters that describe experiment

        scoring_metrics: List of metrics to calculate during evaluation.
            Each metric has `score(...)` method, arguments for this method
            are taken from the `task` output, check the signature
            of the `score` method in metrics that you need to find out which keys
            are mandatory in `task`-returned dictionary.

        verbose: an integer value that controls evaluation output logs such as summary and tqdm progress bar.
            0 - no outputs, 1 - outputs are enabled (default).

        task_threads: amount of thread workers to run tasks. If set to 1, no additional
            threads are created, all tasks executed in the current thread sequentially.
            are executed sequentially in the current thread.
            Use more than 1 worker if your task object is compatible with sharing across threads.
    """
    client = opik_client.get_client_cached()
    start_time = time.time()

    test_results = tasks_scorer.run(
        client=client,
        dataset_=dataset,
        task=task,
        scoring_metrics=scoring_metrics,
        workers=task_threads,
        verbose=verbose,
    )

    total_time = time.time() - start_time

    if verbose == 1:
        report.display_experiment_results(dataset.name, total_time, test_results)

    scores_logger.log_scores(client=client, test_results=test_results)

    experiment = client.create_experiment(
        name=experiment_name,
        dataset_name=dataset.name,
        experiment_config=experiment_config,
    )
    experiment_items = [
        experiment_item.ExperimentItem(
            dataset_item_id=result.test_case.dataset_item_id,
            trace_id=result.test_case.trace_id,
        )
        for result in test_results
    ]

    experiment.insert(experiment_items=experiment_items)

    client.flush()

    evaluation_result_ = evaluation_result.EvaluationResult(
        experiment_id=experiment.id,
        experiment_name=experiment.name,
        test_results=test_results,
    )
    return evaluation_result_



import logging
import dataclasses
import datetime as dt
from enum import Enum
from pathlib import PurePath
from types import GeneratorType
from typing import Any

import numpy as np

import opik.rest_api.core.datetime_utils as datetime_utils

LOGGER = logging.getLogger(__name__)


def jsonable_encoder(obj: Any) -> Any:
    """
    This is a modified version of the serializer generated by Fern in rest_api.core.jsonable_encoder.
    The code is simplified to serialize complex objects into a textual representation.
    """
    try:
        if dataclasses.is_dataclass(obj):
            obj_dict = obj.__dict__
            return jsonable_encoder(obj_dict)
        if isinstance(obj, Enum):
            return jsonable_encoder(obj.value)
        if isinstance(obj, PurePath):
            return str(obj)
        if isinstance(obj, (str, int, float, type(None))):
            return obj
        if isinstance(obj, dt.datetime):
            return datetime_utils.serialize_datetime(obj)
        if isinstance(obj, dt.date):
            return str(obj)
        if isinstance(obj, dict):
            encoded_dict = {}
            allowed_keys = set(obj.keys())
            for key, value in obj.items():
                if key in allowed_keys:
                    encoded_key = jsonable_encoder(key)
                    encoded_value = jsonable_encoder(value)
                    encoded_dict[encoded_key] = encoded_value
            return encoded_dict
        if isinstance(obj, (list, set, frozenset, GeneratorType, tuple)):
            encoded_list = []
            for item in obj:
                encoded_list.append(jsonable_encoder(item))
            return encoded_list

        if isinstance(obj, np.ndarray):
            return jsonable_encoder(obj.tolist())

        if hasattr(obj, "to_string"):  # langchain internal data objects
            try:
                return jsonable_encoder(obj.to_string())
            except Exception:
                pass

    except Exception:
        LOGGER.debug("Failed to serialize object.", exc_info=True)

    data = str(obj)

    return data



"""CLI tool for Opik."""

from importlib import metadata

import click
import questionary

from opik.configurator import configure as opik_configure

__version__: str = "0.0.0+dev"
if __package__:
    __version__ = metadata.version(__package__)

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, *("--version", "-v"))
def cli() -> None:
    """CLI tool for Opik."""


@cli.command(context_settings={"ignore_unknown_options": True})
@click.option(
    "--use_local",
    is_flag=True,
    default=False,
    help="Flag to configure the Opik Python SDK for local Opik deployments.",
)
def configure(use_local: bool) -> None:
    """
    Create a configuration file for the Opik Python SDK, if a configuration file already exists, it will be overwritten. This is also available as a function in the Python SDK.
    """
    if use_local:
        opik_configure.configure(use_local=True, force=True)
    else:
        deployment_type_choice = questionary.select(
            "Which Opik deployment do you want to log your traces to?",
            choices=["Opik Cloud", "Local deployment"],
        ).unsafe_ask()

        if deployment_type_choice == "Opik Cloud":
            opik_configure.configure(use_local=False, force=True)
        else:
            opik_configure.configure(use_local=True, force=True)


if __name__ == "__main__":
    cli()



from typing import List, Callable, cast
from pytest import Item

import json

from opik.api_objects import opik_client
from opik.api_objects.experiment import experiment_item
from opik.api_objects.dataset import dataset_item
from opik.api_objects import helpers

from opik import datetime_helpers
from . import test_runs_storage, test_run_content


def get_dataset_item_id_finder(
    existing_dataset_items: List[dataset_item.DatasetItem],
) -> Callable[[test_run_content.TestRunContent], str]:
    """
    Returns a callable that takes TestRunContent as input and checks
    if there is already a dataset item with such a content (input, expected_output and metadata).
    """
    dataset_item_content_to_ids = {
        json.dumps(dataset_item.get_content(), sort_keys=True): dataset_item.id
        for dataset_item in existing_dataset_items
    }

    def callback(content: test_run_content.TestRunContent) -> str:
        content_str = json.dumps(content.__dict__, sort_keys=True)
        result = dataset_item_content_to_ids.get(content_str, None)
        return cast(str, result)

    return callback


def run(client: opik_client.Opik, test_items: List[Item]) -> None:
    experiment_name = f"Test-Suite-{datetime_helpers.local_timestamp()}"

    try:
        dataset = client.get_dataset("tests")
    except Exception:
        dataset = client.create_dataset("tests")

    dataset_items = dataset.get_all_items()
    dataset_item_id_finder = get_dataset_item_id_finder(
        existing_dataset_items=dataset_items
    )

    experiment = client.create_experiment(name=experiment_name, dataset_name="tests")

    experiment_items: List[experiment_item.ExperimentItem] = []
    dataset_items_to_create: List[dataset_item.DatasetItem] = []

    for test_item in test_items:
        test_run_content = test_runs_storage.TEST_RUNS_CONTENTS[test_item.nodeid]
        test_run_trace_id = test_runs_storage.TEST_RUNS_TO_TRACE_DATA[
            test_item.nodeid
        ].id

        dataset_item_id = dataset_item_id_finder(test_run_content)

        if dataset_item_id is None:
            dataset_item_id = helpers.generate_id()
            dataset_item_ = dataset_item.DatasetItem(
                id=dataset_item_id,
                input=test_run_content.input,
                expected_output=test_run_content.expected_output,
                metadata=test_run_content.metadata,
            )
            dataset_items_to_create.append(dataset_item_)

        experiment_items.append(
            experiment_item.ExperimentItem(
                dataset_item_id=dataset_item_id,
                trace_id=test_run_trace_id,
            )
        )

    dataset.insert(items=dataset_items_to_create)
    experiment.insert(experiment_items=experiment_items)
    client.flush()



import datetime
import uuid_extensions

from typing import List, Any

from .. import datetime_helpers
from typing import Optional


def generate_id() -> str:
    return str(uuid_extensions.uuid7())


def datetime_to_iso8601_if_not_None(
    value: Optional[datetime.datetime],
) -> Optional[str]:
    if value is None:
        return None

    return datetime_helpers.datetime_to_iso8601(value)


def list_to_batches(items: List[Any], batch_size: int) -> List[List[Any]]:
    return [items[i : i + batch_size] for i in range(0, len(items), batch_size)]



from .decorator.tracker import track, flush_tracker
from .api_objects.opik_client import Opik
from .api_objects.trace import Trace
from .api_objects.span import Span
from .api_objects.dataset.dataset_item import DatasetItem
from .api_objects.dataset import Dataset
from . import _logging
from .configurator.configure import configure
from . import package_version
from .plugins.pytest.decorator import llm_unit
from .evaluation import evaluate

_logging.setup()

__version__ = package_version.VERSION
__all__ = [
    "__version__",
    "evaluate",
    "track",
    "flush_tracker",
    "Opik",
    "Trace",
    "Span",
    "DatasetItem",
    "Dataset",
    "llm_unit",
    "configure",
]



import logging
import sys

LOGGER = logging.getLogger(__name__)


def is_interactive() -> bool:
    """
    Determines if the current environment is interactive.

    Returns:
        bool: True if the environment is either running in a terminal,
              a Jupyter notebook, an IPython environment, or Google Colab.
              False otherwise.
    """
    return (
        sys.stdin.isatty()
        or _in_jupyter_environment()
        or _in_ipython_environment()
        or _in_colab_environment()
    )


def _in_jupyter_environment() -> bool:
    """
    Determine if the current environment is a Jupyter notebook.

    Returns:
        bool: True if running in a Jupyter notebook environment, otherwise False.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    if ipy is None or not hasattr(ipy, "kernel"):
        return False
    else:
        return True


def _in_ipython_environment() -> bool:
    """
    Determines if the current environment is an IPython environment.

    Returns:
        bool: True if the code is running in an IPython environment, False otherwise.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    if ipy is None:
        return False
    else:
        return True


def _in_colab_environment() -> bool:
    """
    Determines if the code is running within a Google Colab environment.

    Returns:
        bool: True if running in Google Colab, False otherwise.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    return "google.colab" in str(ipy)


def ask_user_for_approval(message: str) -> bool:
    """
    Prompt the user with a message for approval (Y/Yes/N/No).

    Args:
        message (str): The message to display to the user.

    Returns:
        bool: True if the user approves (Y/Yes/empty input), False if the user disapproves (N/No).

    Logs:
        Error when the user input is not recognized.
    """
    while True:
        users_choice = input(message).strip().upper()
        if users_choice in ("Y", "YES", ""):
            return True
        if users_choice in ("N", "NO"):
            return False
        LOGGER.error("Wrong choice. Please try again.")



from typing import List, TypedDict

HALLUCINATION_VERDICT = "hallucinated"
FACTUAL_VERDICT = "factual"

VERDICT_KEY = "verdict"
REASON_KEY = "reason"


class FewShotExampleHallucination(TypedDict):
    title: str
    input: str
    context: List[str]
    output: str
    verdict: str
    reason: str


def generate_query(
    input: str,
    output: str,
    context: List[str],
    few_shot_examples: List[FewShotExampleHallucination],
) -> str:
    if len(few_shot_examples) == 0:
        examples_str = ""
    else:
        examples_str = "\n\nEXAMPLES:\n\n".join(
            [
                f"<example>\n"
                f"Input: {example['input']}\n"
                f"Context: {example['context']}\n"
                f"Output: {example['output']}\n\n"
                f"{{\"{VERDICT_KEY}\": \"{example['verdict']}\", \"{REASON_KEY}\": \"{example['reason']}\"}}\n"
                f"</example>"
                for i, example in enumerate(few_shot_examples)
            ]
        )

    return f"""You are an expert judge tasked with evaluating the faithfulness of an AI-generated answer to the given context. Analyze the provided INPUT, CONTEXT, and OUTPUT to determine if the OUTPUT contains any hallucinations or unfaithful information.

Guidelines:
1. The OUTPUT must not introduce new information beyond what's provided in the CONTEXT.
2. The OUTPUT must not contradict any information given in the CONTEXT.
3. Ignore the INPUT when evaluating faithfulness; it's provided for context only.
4. Consider partial hallucinations where some information is correct but other parts are not.
5. Pay close attention to the subject of statements. Ensure that attributes, actions, or dates are correctly associated with the right entities (e.g., a person vs. a TV show they star in).
6. Be vigilant for subtle misattributions or conflations of information, even if the date or other details are correct.
7. Check that the OUTPUT doesn't oversimplify or generalize information in a way that changes its meaning or accuracy.

Verdict options:
- "{FACTUAL_VERDICT}": The OUTPUT is entirely faithful to the CONTEXT.
- "{HALLUCINATION_VERDICT}": The OUTPUT contains hallucinations or unfaithful information.

{examples_str}

INPUT (for context only, not to be used for faithfulness evaluation):
{input}

CONTEXT:
{context}

OUTPUT:
{output}

Provide your verdict in JSON format:
{{
    "{VERDICT_KEY}": <your verdict>,
    "{REASON_KEY}": [
        <list your reasoning as bullet points>
    ]
}}
    """



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from ..core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1
from .dataset_item_compare import DatasetItemCompare


class DatasetItemPageCompare(pydantic_v1.BaseModel):
    content: typing.Optional[typing.List[DatasetItemCompare]] = None
    page: typing.Optional[int] = None
    size: typing.Optional[int] = None
    total: typing.Optional[int] = None

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {
            "by_alias": True,
            "exclude_unset": True,
            **kwargs,
        }
        kwargs_with_defaults_exclude_none: typing.Any = {
            "by_alias": True,
            "exclude_none": True,
            **kwargs,
        }

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset),
            super().dict(**kwargs_with_defaults_exclude_none),
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



import queue
import threading
from typing import Any, List, Optional

from . import messages, queue_consumer
from .. import synchronization
from .batching import batch_manager


class Streamer:
    def __init__(
        self,
        message_queue: "queue.Queue[Any]",
        queue_consumers: List[queue_consumer.QueueConsumer],
        batch_manager: Optional[batch_manager.BatchManager],
    ) -> None:
        self._lock = threading.RLock()
        self._message_queue = message_queue
        self._queue_consumers = queue_consumers
        self._batch_manager = batch_manager

        self._drain = False

        self._start_queue_consumers()

        if self._batch_manager is not None:
            self._batch_manager.start()

    def put(self, message: messages.BaseMessage) -> None:
        with self._lock:
            if self._drain:
                return

            if (
                self._batch_manager is not None
                and self._batch_manager.message_supports_batching(message)
            ):
                self._batch_manager.process_message(message)
            else:
                self._message_queue.put(message)

    def close(self, timeout: Optional[int]) -> bool:
        """
        Stops data sending threads
        """
        with self._lock:
            self._drain = True

        if self._batch_manager is not None:
            self._batch_manager.stop()  # stopping causes adding remaining batch messages to the queue

        self.flush(timeout)
        self._close_queue_consumers()

        return self._message_queue.empty()

    def flush(self, timeout: Optional[int]) -> None:
        if self._batch_manager is not None:
            self._batch_manager.flush()

        synchronization.wait_for_done(
            check_function=lambda: (
                self.workers_waiting()
                and self._message_queue.empty()
                and (self._batch_manager is None or self._batch_manager.is_empty())
            ),
            timeout=timeout,
            sleep_time=0.1,
        )

    def workers_waiting(self) -> bool:
        return all([consumer.waiting for consumer in self._queue_consumers])

    def _start_queue_consumers(self) -> None:
        for consumer in self._queue_consumers:
            consumer.start()

    def _close_queue_consumers(self) -> None:
        for consumer in self._queue_consumers:
            consumer.close()



from typing import Dict, Any, Optional

import dataclasses


@dataclasses.dataclass(frozen=True, eq=True)
class TestRunContent:
    input: Dict[str, Any]
    """The input data for the test."""
    expected_output: Optional[Dict[str, Any]] = None
    """The expected output for the model used in test."""
    metadata: Optional[Dict[str, Any]] = None
    """Additional metadata associated with the test run."""



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Literal, Set

from langchain_core.tracers import BaseTracer

from opik.api_objects import opik_client, span, trace
from opik import dict_utils
from opik import opik_context

from . import openai_run_helpers


if TYPE_CHECKING:
    from uuid import UUID

    from langchain_core.tracers.schemas import Run


def _get_span_type(run: "Run") -> Literal["llm", "tool", "general"]:
    if run.run_type in ["llm", "tool"]:
        return run.run_type  # type: ignore[no-any-return]

    return "general"


class OpikTracer(BaseTracer):
    """Opik Tracer."""

    def __init__(
        self,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the Opik Tracer.

        Args:
            tags: List of tags to be applied to each trace logged by the tracer.
            metadata: Additional metadata for each trace logged by the tracer.
        """
        super().__init__(**kwargs)
        self._trace_default_metadata = metadata if metadata is not None else {}
        self._trace_default_tags = tags

        self._span_data_map: Dict["UUID", span.SpanData] = {}
        """Map from run id to span data."""

        self._created_traces_data_map: Dict["UUID", trace.TraceData] = {}
        """Map from run id to trace data."""

        self._created_traces: List[trace.Trace] = []

        self._externally_created_traces_ids: Set[str] = set()

        self._opik_client = opik_client.get_client_cached()

    def _persist_run(self, run: "Run") -> None:
        run_dict: Dict[str, Any] = run.dict()

        span_data = self._span_data_map[run.id]
        span_data.init_end_time().update(output=run_dict["outputs"])
        self._opik_client.span(**span_data.__dict__)

        if span_data.trace_id not in self._externally_created_traces_ids:
            trace_data = self._created_traces_data_map[run.id]
            trace_data.init_end_time().update(output=run_dict["outputs"])
            trace_ = self._opik_client.trace(**trace_data.__dict__)
            self._created_traces.append(trace_)

    def _process_start_trace(self, run: "Run") -> None:
        run_dict: Dict[str, Any] = run.dict()
        if not run.parent_run_id:
            # This is the first run for the chain.
            self._track_root_run(run_dict)
        else:
            parent_span_data = self._span_data_map[run.parent_run_id]
            span_data = span.SpanData(
                trace_id=parent_span_data.trace_id,
                parent_span_id=parent_span_data.id,
                input=run_dict["inputs"],
                metadata=run_dict["extra"],
                name=run.name,
                type=_get_span_type(run),
            )

            self._span_data_map[run.id] = span_data

            if span_data.trace_id not in self._externally_created_traces_ids:
                self._created_traces_data_map[run.id] = self._created_traces_data_map[
                    run.parent_run_id
                ]

    def _track_root_run(self, run_dict: Dict[str, Any]) -> None:
        run_metadata = run_dict["extra"].get("metadata", {})
        root_metadata = dict_utils.deepmerge(self._trace_default_metadata, run_metadata)
        current_span_data = opik_context.get_current_span_data()
        if current_span_data is not None:
            self._attach_span_to_existing_span(
                run_dict=run_dict,
                current_span_data=current_span_data,
                root_metadata=root_metadata,
            )
            return

        current_trace_data = opik_context.get_current_trace_data()
        if current_trace_data is not None:
            self._attach_span_to_existing_trace(
                run_dict=run_dict,
                current_trace_data=current_trace_data,
                root_metadata=root_metadata,
            )
            return

        self._initialize_span_and_trace_from_scratch(
            run_dict=run_dict, root_metadata=root_metadata
        )

    def _initialize_span_and_trace_from_scratch(
        self, run_dict: Dict[str, Any], root_metadata: Dict[str, Any]
    ) -> None:
        trace_data = trace.TraceData(
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
        )

        self._created_traces_data_map[run_dict["id"]] = trace_data

        span_ = span.SpanData(
            trace_id=trace_data.id,
            parent_span_id=None,
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
        )

        self._span_data_map[run_dict["id"]] = span_

    def _attach_span_to_existing_span(
        self,
        run_dict: Dict[str, Any],
        current_span_data: span.SpanData,
        root_metadata: Dict[str, Any],
    ) -> None:
        span_data = span.SpanData(
            trace_id=current_span_data.trace_id,
            parent_span_id=current_span_data.id,
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
        )
        self._span_data_map[run_dict["id"]] = span_data
        self._externally_created_traces_ids.add(span_data.trace_id)

    def _attach_span_to_existing_trace(
        self,
        run_dict: Dict[str, Any],
        current_trace_data: trace.TraceData,
        root_metadata: Dict[str, Any],
    ) -> None:
        span_data = span.SpanData(
            trace_id=current_trace_data.id,
            parent_span_id=None,
            name=run_dict["name"],
            input=run_dict["inputs"],
            metadata=root_metadata,
            tags=self._trace_default_tags,
        )
        self._span_data_map[run_dict["id"]] = span_data
        self._externally_created_traces_ids.add(current_trace_data.id)

    def _process_end_trace(self, run: "Run") -> None:
        run_dict: Dict[str, Any] = run.dict()
        if not run.parent_run_id:
            pass
            # Langchain will call _persist_run for us
        else:
            span_data = self._span_data_map[run.id]
            if openai_run_helpers.is_openai_run(run):
                usage = openai_run_helpers.try_get_token_usage(run_dict)
            else:
                usage = None

            span_data.init_end_time().update(
                output=run_dict["outputs"],
                usage=usage,
            )
            self._opik_client.span(**span_data.__dict__)

    def flush(self) -> None:
        """
        Flush to ensure all data is sent to the Opik server.
        """
        self._opik_client.flush()

    def created_traces(self) -> List[trace.Trace]:
        """
        Get a list of traces created by OpikTracer.

        Returns:
            List[Trace]: A list of traces.
        """
        return self._created_traces

    def _on_llm_start(self, run: "Run") -> None:
        """Process the LLM Run upon start."""
        self._process_start_trace(run)

    def _on_llm_end(self, run: "Run") -> None:
        """Process the LLM Run."""
        self._process_end_trace(run)

    def _on_llm_error(self, run: "Run") -> None:
        """Process the LLM Run upon error."""
        self._process_end_trace(run)

    def _on_chain_start(self, run: "Run") -> None:
        """Process the Chain Run upon start."""
        self._process_start_trace(run)

    def _on_chain_end(self, run: "Run") -> None:
        """Process the Chain Run."""
        self._process_end_trace(run)

    def _on_chain_error(self, run: "Run") -> None:
        """Process the Chain Run upon error."""
        self._process_end_trace(run)

    def _on_tool_start(self, run: "Run") -> None:
        """Process the Tool Run upon start."""
        self._process_start_trace(run)

    def _on_tool_end(self, run: "Run") -> None:
        """Process the Tool Run."""
        self._process_end_trace(run)

    def _on_tool_error(self, run: "Run") -> None:
        """Process the Tool Run upon error."""
        self._process_end_trace(run)



import logging
from opik import _logging
from typing import List, Any, Generator, TYPE_CHECKING
from opik.types import FeedbackScoreDict

from opik.api_objects import opik_client
from . import test_runs_storage, experiment_runner, summary

import pytest

if TYPE_CHECKING:
    import _pytest.terminal

LOGGER = logging.getLogger(__name__)


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item: "pytest.Item") -> Generator:
    """
    Write the results of each test in the session
    """
    outcome = yield

    try:
        report = outcome.get_result()
        if report.when == "call":
            item.report = report
    except Exception:
        LOGGER.debug(
            "Unexpected failure during opik pytest_runtest_makereport hook",
            exc_info=True,
        )


@_logging.convert_exception_to_log_message(
    "Unexpected failure during opik pytest_sessionfinish hook",
    logger=LOGGER,
    exc_info=True,
    logging_level=logging.ERROR,
)
def pytest_sessionfinish(session: "pytest.Session", exitstatus: Any) -> None:
    llm_test_items: List["pytest.Item"] = [
        test_item
        for test_item in session.items
        if test_item.nodeid in test_runs_storage.LLM_UNIT_TEST_RUNS
    ]
    if len(llm_test_items) == 0:
        return

    try:
        traces_feedback_scores: List[FeedbackScoreDict] = []

        for item in llm_test_items:
            report: "pytest.TestReport" = item.report
            trace_id = test_runs_storage.TEST_RUNS_TO_TRACE_DATA[item.nodeid].id
            traces_feedback_scores.append(
                {"id": trace_id, "name": "Passed", "value": report.passed}
            )

        client = opik_client.get_client_cached()
        client.log_traces_feedback_scores(traces_feedback_scores)

        experiment_runner.run(client=client, test_items=llm_test_items)
        client.flush()
    except Exception:
        LOGGER.error(
            "Unexpected exception occured while trying to log LLM unit tests experiment results",
            exc_info=True,
        )


@_logging.convert_exception_to_log_message(
    "Unexpected exception occured while trying to print LLM unit tests summary",
    logger=LOGGER,
    exc_info=True,
    logging_level=logging.DEBUG,
)
def pytest_terminal_summary(
    terminalreporter: "_pytest.terminal.TerminalReporter",
) -> None:
    reports: List[pytest.TestReport] = terminalreporter.stats.get(
        "passed", []
    ) + terminalreporter.stats.get("failed", [])

    llm_reports = [
        report
        for report in reports
        if report.nodeid in test_runs_storage.LLM_UNIT_TEST_RUNS
    ]

    if len(llm_reports) == 0:
        return

    summary.print(llm_reports)



import datetime
import dataclasses
import logging

from typing import Optional, Any, List, Dict
from ..types import SpanType, UsageDict, DistributedTraceHeadersDict, FeedbackScoreDict

from ..message_processing import streamer, messages
from .. import datetime_helpers
from . import helpers, validation_helpers, constants

LOGGER = logging.getLogger(__name__)


class Span:
    def __init__(
        self,
        id: str,
        trace_id: str,
        project_name: str,
        message_streamer: streamer.Streamer,
        parent_span_id: Optional[str] = None,
    ):
        """
        A Span object. This object should not be created directly, instead use the `span` method of a Trace (:func:`opik.Opik.span`) or another Span (:meth:`opik.Span.span`).
        """
        self.id = id
        self.trace_id = trace_id
        self.parent_span_id = parent_span_id
        self._streamer = message_streamer
        self._project_name = project_name

    def end(
        self,
        end_time: Optional[datetime.datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        usage: Optional[UsageDict] = None,
    ) -> None:
        """
        End the span and update its attributes.

        This method is similar to the `update` method, but it automatically computes
        the end time if not provided.

        Args:
            end_time: The end time of the span. If not provided, the current time will be used.
            metadata: Additional metadata to be associated with the span.
            input: The input data for the span.
            output: The output data for the span.
            tags: A list of tags to be associated with the span.
            usage: Usage information for the span.

        Returns:
            None
        """
        end_time = (
            end_time if end_time is not None else datetime_helpers.local_timestamp()
        )

        self.update(
            end_time=end_time,
            metadata=metadata,
            input=input,
            output=output,
            tags=tags,
            usage=usage,
        )

    def update(
        self,
        end_time: Optional[datetime.datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        usage: Optional[UsageDict] = None,
    ) -> None:
        """
        Update the span attributes.

        Args:
            end_time: The end time of the span.
            metadata: Additional metadata to be associated with the span.
            input: The input data for the span.
            output: The output data for the span.
            tags: A list of tags to be associated with the span.
            usage: Usage information for the span.

        Returns:
            None
        """
        parsed_usage = validation_helpers.validate_and_parse_usage(usage, LOGGER)
        if parsed_usage.full_usage is not None:
            metadata = (
                {"usage": parsed_usage.full_usage}
                if metadata is None
                else {"usage": parsed_usage.full_usage, **metadata}
            )

        end_span_message = messages.UpdateSpanMessage(
            span_id=self.id,
            trace_id=self.trace_id,
            parent_span_id=self.parent_span_id,
            project_name=self._project_name,
            end_time=end_time,
            metadata=metadata,
            input=input,
            output=output,
            tags=tags,
            usage=parsed_usage.supported_usage,
        )
        self._streamer.put(end_span_message)

    def span(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        type: SpanType = "general",
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        usage: Optional[UsageDict] = None,
    ) -> "Span":
        """
        Create a new child span within the current span.

        Args:
            id: The ID of the span, should be in UUIDv7 format. If not provided, a new ID will be generated.
            name: The name of the span.
            type: The type of the span. Defaults to "general".
            start_time: The start time of the span. If not provided, current time will be used.
            end_time: The end time of the span.
            metadata: Additional metadata to be associated with the span.
            input: The input data for the span.
            output: The output data for the span.
            tags: A list of tags to be associated with the span.
            usage: Usage information for the span.

        Returns:
            Span: The created child span object.
        """
        span_id = id if id is not None else helpers.generate_id()
        start_time = (
            start_time if start_time is not None else datetime_helpers.local_timestamp()
        )
        parsed_usage = validation_helpers.validate_and_parse_usage(usage, LOGGER)
        if parsed_usage.full_usage is not None:
            metadata = (
                {"usage": parsed_usage.full_usage}
                if metadata is None
                else {"usage": parsed_usage.full_usage, **metadata}
            )

        create_span_message = messages.CreateSpanMessage(
            span_id=span_id,
            trace_id=self.trace_id,
            project_name=self._project_name,
            parent_span_id=self.id,
            name=name,
            type=type,
            start_time=start_time,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            tags=tags,
            usage=parsed_usage.supported_usage,
        )
        self._streamer.put(create_span_message)

        return Span(
            id=span_id,
            parent_span_id=self.id,
            trace_id=self.trace_id,
            message_streamer=self._streamer,
            project_name=self._project_name,
        )

    def log_feedback_score(
        self,
        name: str,
        value: float,
        category_name: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> None:
        """
        Log a feedback score for the span.

        Args:
            name: The name of the feedback score.
            value: The value of the feedback score.
            category_name: The category name for the feedback score.
            reason: The reason for the feedback score.

        Returns:
            None
        """
        add_span_feedback_batch_message = messages.AddSpanFeedbackScoresBatchMessage(
            batch=[
                messages.FeedbackScoreMessage(
                    id=self.id,
                    name=name,
                    value=value,
                    category_name=category_name,
                    reason=reason,
                    source=constants.FEEDBACK_SCORE_SOURCE_SDK,
                    project_name=self._project_name,
                )
            ],
        )

        self._streamer.put(add_span_feedback_batch_message)

    def get_distributed_trace_headers(self) -> DistributedTraceHeadersDict:
        """
        Returns headers dictionary to be passed into tracked
        function on remote node.
        """
        return {"opik_parent_span_id": self.id, "opik_trace_id": self.trace_id}


@dataclasses.dataclass
class SpanData:
    """
    The SpanData object is returned when calling :func:`opik.opik_context.get_current_span_data` from a tracked function.
    """

    trace_id: str
    id: str = dataclasses.field(default_factory=helpers.generate_id)
    parent_span_id: Optional[str] = None
    name: Optional[str] = None
    type: SpanType = "general"
    start_time: Optional[datetime.datetime] = dataclasses.field(
        default_factory=datetime_helpers.local_timestamp
    )
    end_time: Optional[datetime.datetime] = None
    metadata: Optional[Dict[str, Any]] = None
    input: Optional[Dict[str, Any]] = None
    output: Optional[Dict[str, Any]] = None
    tags: Optional[List[str]] = None
    usage: Optional[UsageDict] = None
    feedback_scores: Optional[List[FeedbackScoreDict]] = None

    def update(self, **new_data: Any) -> "SpanData":
        for key, value in new_data.items():
            if value is not None:
                if key in self.__dict__:
                    self.__dict__[key] = value
                else:
                    LOGGER.debug(
                        "An attempt to update span with parameter name it doesn't have: %s",
                        key,
                    )

        return self

    def init_end_time(self) -> "SpanData":
        self.end_time = datetime_helpers.local_timestamp()

        return self



from typing import List

import dataclasses

from . import test_result


@dataclasses.dataclass
class EvaluationResult:
    experiment_id: str
    experiment_name: str
    test_results: List[test_result.TestResult]



import copy
import logging
from typing import Any, Dict, Mapping, Optional

from . import logging_messages

LOGGER = logging.getLogger(__name__)


def deepmerge(
    dict1: Dict[str, Any], dict2: Dict[str, Any], max_depth: int = 10
) -> Dict[str, Any]:
    merged = copy.deepcopy(dict1)

    for key, value in dict2.items():
        if (
            key in merged
            and _is_dict(merged[key])
            and _is_dict(value)
            and max_depth > 0
        ):
            merged[key] = deepmerge(merged[key], value, max_depth=max_depth - 1)
        else:
            if key in merged:
                LOGGER.debug(
                    logging_messages.METADATA_KEY_COLLISION_DURING_DEEPMERGE,
                    key,
                    merged[key],
                    value,
                )
            merged[key] = value

    return merged


def remove_none_from_dict(original: Mapping[str, Optional[Any]]) -> Mapping[str, Any]:
    new: Dict[str, Any] = {}

    for key, value in original.items():
        if value is not None:
            new[key] = value

    return new


def _is_dict(item: Any) -> bool:
    return isinstance(item, dict)



import logging
from typing import Generator, Any, AsyncGenerator, List, Optional, Callable
from opik.api_objects import trace, span
from opik.decorator import generator_wrappers
from openai.types.chat import chat_completion_chunk
import openai

LOGGER = logging.getLogger(__name__)


def wrap_sync_stream(
    generator: openai.Stream,
    capture_output: bool,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    generations_aggregator: Callable[[List[Any]], Any],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> Generator[Any, None, None]:
    items: List[chat_completion_chunk.ChatCompletionChunk] = []

    try:
        for item in generator:
            items.append(item)

            yield item

    finally:
        output = generations_aggregator(items)
        finally_callback(
            output=output,
            generators_span_to_end=span_to_end,
            generators_trace_to_end=trace_to_end,
            capture_output=capture_output,
        )


async def wrap_async_stream(
    generator: openai.AsyncStream,
    capture_output: bool,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    generations_aggregator: Callable[[List[Any]], Any],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> AsyncGenerator[Any, None]:
    items: List[chat_completion_chunk.ChatCompletionChunk] = []

    try:
        async for item in generator:
            items.append(item)

            yield item

    finally:
        output = generations_aggregator(items)
        finally_callback(
            output=output,
            generators_span_to_end=span_to_end,
            generators_trace_to_end=trace_to_end,
            capture_output=capture_output,
        )



from typing import List, TypedDict

VERDICT_KEY = "context_recall_score"
REASON_KEY = "reason"


class FewShotExampleContextRecall(TypedDict):
    title: str
    input: str
    expected_output: str
    context: str
    output: str
    context_recall_score: float
    reason: str


FEW_SHOT_EXAMPLES: List[FewShotExampleContextRecall] = [
    {
        "title": "Low ContextRecall Score",
        "input": "Provide the name of the capital of a European country.",
        "expected_output": "Paris.",
        "context": "The user is specifically asking about the capital city of the country that hosts the Eiffel Tower.",
        "output": "Berlin.",
        "context_recall_score": 0.2,
        "reason": "The LLM's response 'Berlin' is incorrect. The context specifically refers to a country known for the Eiffel Tower, which is a landmark in France, not Germany. The response fails to address this critical context and provides the wrong capital.",
    },
    {
        "title": "Medium ContextRecall Score",
        "input": "Provide the name of the capital of a European country.",
        "expected_output": "Paris.",
        "context": "The user is specifically asking about the capital city of the country that hosts the Eiffel Tower.",
        "output": "Marseille.",
        "context_recall_score": 0.5,
        "reason": "The LLM's response 'Marseille' is partially correct because it identifies a major city in France. However, it fails to recognize 'Paris' as the capital, especially within the context of the Eiffel Tower, which is located in Paris.",
    },
    {
        "title": "High ContextRecall Score",
        "input": "Provide the name of the capital of a European country.",
        "expected_output": "Paris.",
        "context": "The user is specifically asking about the capital city of the country that hosts the Eiffel Tower.",
        "output": "Paris, the capital of France, is where the Eiffel Tower is located.",
        "context_recall_score": 0.9,
        "reason": "The LLM's response is highly accurate, correctly identifying 'Paris' as the capital of France and incorporating the reference to the Eiffel Tower mentioned in the context. The response is comprehensive but slightly more detailed than necessary, preventing a perfect score.",
    },
]


def generate_query(
    input: str,
    output: str,
    expected_output: str,
    context: List[str],
    few_shot_examples: List[FewShotExampleContextRecall],
) -> str:
    examples_str = "\n\n".join(
        [
            f"#### Example {i+1}: {example['title']}\n\n"
            f"- **Input:** \"{example['input']}\"\n"
            f"- **Output:** \"{example['output']}\"\n"
            f"- **Expected Output:** \"{example['expected_output']}\"\n"
            f"- **Context:** \"{example['context']}\"\n"
            f"- **Result:**\n"
            f"  ```json\n"
            f"  {{\n"
            f"    \"{VERDICT_KEY}\": {example['context_recall_score']},\n"
            f"    \"{REASON_KEY}\": \"{example['reason']}\"\n"
            f"  }}\n"
            f"  ```"
            for i, example in enumerate(few_shot_examples)
        ]
    )

    return f"""YOU ARE AN EXPERT AI METRIC EVALUATOR SPECIALIZING IN CONTEXTUAL UNDERSTANDING AND RESPONSE ACCURACY.
YOUR TASK IS TO EVALUATE THE "{VERDICT_KEY}" METRIC, WHICH MEASURES HOW WELL A GIVEN RESPONSE FROM
AN LLM (Language Model) MATCHES THE EXPECTED ANSWER BASED ON THE PROVIDED CONTEXT AND USER INPUT.

###INSTRUCTIONS###

1. **Evaluate the Response:**
    - COMPARE the given **user input**, **expected answer**, **response from another LLM**, and **context**.
    - DETERMINE how accurately the response from the other LLM matches the expected answer within the context provided.

2. **Score Assignment:**
    - ASSIGN a **{VERDICT_KEY}** score on a scale from **0.0 to 1.0**:
        - **0.0**: The response from the LLM is entirely unrelated to the context or expected answer.
        - **0.1 - 0.3**: The response is minimally relevant but misses key points or context.
        - **0.4 - 0.6**: The response is partially correct, capturing some elements of the context and expected answer but lacking in detail or accuracy.
        - **0.7 - 0.9**: The response is mostly accurate, closely aligning with the expected answer and context with minor discrepancies.
        - **1.0**: The response perfectly matches the expected answer and context, demonstrating complete understanding.

3. **Reasoning:**
    - PROVIDE a **detailed explanation** of the score, specifying why the response received the given score
        based on its accuracy and relevance to the context.

4. **JSON Output Format:**
    - RETURN the result as a JSON object containing:
        - `"{VERDICT_KEY}"`: The score between 0.0 and 1.0.
        - `"{REASON_KEY}"`: A detailed explanation of the score.

###CHAIN OF THOUGHTS###

1. **Understand the Context:**
    1.1. Analyze the context provided.
    1.2. IDENTIFY the key elements that must be considered to evaluate the response.

2. **Compare the Expected Answer and LLM Response:**
    2.1. CHECK the LLM's response against the expected answer.
    2.2. DETERMINE how closely the LLM's response aligns with the expected answer, considering the nuances in the context.

3. **Assign a Score:**
    3.1. REFER to the scoring scale.
    3.2. ASSIGN a score that reflects the accuracy of the response.

4. **Explain the Score:**
    4.1. PROVIDE a clear and detailed explanation.
    4.2. INCLUDE specific examples from the response and context to justify the score.

###WHAT NOT TO DO###

- **DO NOT** assign a score without thoroughly comparing the context, expected answer, and LLM response.
- **DO NOT** provide vague or non-specific reasoning for the score.
- **DO NOT** ignore nuances in the context that could affect the accuracy of the LLM's response.
- **DO NOT** assign scores outside the 0.0 to 1.0 range.
- **DO NOT** return any output format other than JSON.

###FEW-SHOT EXAMPLES###

{examples_str}

###INPUTS:###
***
Input:
{input}

Output:
{output}

Expected Output:
{expected_output}

Context:
{context}
***
    """



from typing import Optional, Dict, Any, List
from opik.types import UsageDict, DistributedTraceHeadersDict, FeedbackScoreDict
from opik.api_objects import span, trace

from . import context_storage, exceptions


def get_current_span_data() -> Optional[span.SpanData]:
    """
    Returns current span created by track() decorator or None if no span was found.
    """
    span_data = context_storage.top_span_data()
    if span_data is None:
        return None

    return span.SpanData(**span_data.__dict__)


def get_current_trace_data() -> Optional[trace.TraceData]:
    """
    Returns current trace created by track() decorator or None if no trace was found.
    """
    trace_data = context_storage.get_trace_data()
    if trace_data is None:
        return None

    return trace.TraceData(**trace_data.__dict__)


def get_distributed_trace_headers() -> DistributedTraceHeadersDict:
    """
    Returns headers dictionary to be passed into tracked function on remote node.
    Requires an existing span in the context, otherwise raises an error.
    """
    current_span_data = context_storage.top_span_data()

    if current_span_data is None:
        raise Exception("There is no span in the context.")

    return DistributedTraceHeadersDict(
        opik_trace_id=current_span_data.trace_id,
        opik_parent_span_id=current_span_data.id,
    )


def update_current_span(
    name: Optional[str] = None,
    input: Optional[Dict[str, Any]] = None,
    output: Optional[Dict[str, Any]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    usage: Optional[UsageDict] = None,
    feedback_scores: Optional[List[FeedbackScoreDict]] = None,
) -> None:
    """
    Update the current span with the provided parameters. This method is usually called within a tracked function.

    Args:
        name: The name of the span.
        input: The input data of the span.
        output: The output data of the span.
        metadata: The metadata of the span.
        tags: The tags of the span.
        usage: The usage data of the span.
        feedback_scores: The feedback scores of the span.
    """
    new_params = {
        "name": name,
        "input": input,
        "output": output,
        "metadata": metadata,
        "tags": tags,
        "usage": usage,
        "feedback_scores": feedback_scores,
    }
    current_span_data = context_storage.top_span_data()
    if current_span_data is None:
        raise exceptions.OpikException("There is no span in the context.")

    current_span_data.update(**new_params)


def update_current_trace(
    name: Optional[str] = None,
    input: Optional[Dict[str, Any]] = None,
    output: Optional[Dict[str, Any]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    feedback_scores: Optional[List[FeedbackScoreDict]] = None,
) -> None:
    """
    Update the current trace with the provided parameters. This method is usually called within a tracked function.

    Args:
        name: The name of the trace.
        input: The input data of the trace.
        output: The output data of the trace.
        metadata: The metadata of the trace.
        tags: The tags of the trace.
        feedback_scores: The feedback scores of the trace.
    """
    new_params = {
        "name": name,
        "input": input,
        "output": output,
        "metadata": metadata,
        "tags": tags,
        "feedback_scores": feedback_scores,
    }
    current_trace_data = context_storage.get_trace_data()
    if current_trace_data is None:
        raise exceptions.OpikException("There is no trace in the context.")

    current_trace_data.update(**new_params)


__all__ = [
    "get_current_span_data",
    "get_current_trace_data",
    "update_current_span",
    "update_current_trace",
    "get_distributed_trace_headers",
]



from typing import Optional, Any
from . import langchain_chat_model, base_model

DEFAULT_GPT_MODEL_NAME = "gpt-3.5-turbo"


def get(model_name: Optional[str], **model_kwargs: Any) -> base_model.OpikBaseModel:
    if model_name is None:
        model_name = DEFAULT_GPT_MODEL_NAME

    return langchain_chat_model.LangchainChatModel(
        model_name=model_name, **model_kwargs
    )



# This file was auto-generated by Fern from our API Definition.

from .categorical_feedback_definition import CategoricalFeedbackDefinition
from .categorical_feedback_definition_create import CategoricalFeedbackDefinitionCreate
from .categorical_feedback_definition_public import CategoricalFeedbackDefinitionPublic
from .categorical_feedback_definition_update import CategoricalFeedbackDefinitionUpdate
from .categorical_feedback_detail import CategoricalFeedbackDetail
from .categorical_feedback_detail_create import CategoricalFeedbackDetailCreate
from .categorical_feedback_detail_public import CategoricalFeedbackDetailPublic
from .categorical_feedback_detail_update import CategoricalFeedbackDetailUpdate
from .chunked_output_json_node import ChunkedOutputJsonNode
from .chunked_output_json_node_type import ChunkedOutputJsonNodeType
from .dataset import Dataset
from .dataset_item import DatasetItem
from .dataset_item_batch import DatasetItemBatch
from .dataset_item_compare import DatasetItemCompare
from .dataset_item_compare_source import DatasetItemCompareSource
from .dataset_item_page_compare import DatasetItemPageCompare
from .dataset_item_page_public import DatasetItemPagePublic
from .dataset_item_public import DatasetItemPublic
from .dataset_item_public_source import DatasetItemPublicSource
from .dataset_item_source import DatasetItemSource
from .dataset_item_write import DatasetItemWrite
from .dataset_item_write_source import DatasetItemWriteSource
from .dataset_page_public import DatasetPagePublic
from .dataset_public import DatasetPublic
from .delete_feedback_score import DeleteFeedbackScore
from .error_message import ErrorMessage
from .error_message_public import ErrorMessagePublic
from .experiment import Experiment
from .experiment_item import ExperimentItem
from .experiment_item_compare import ExperimentItemCompare
from .experiment_item_public import ExperimentItemPublic
from .experiment_page_public import ExperimentPagePublic
from .experiment_public import ExperimentPublic
from .feedback import Feedback, Feedback_Categorical, Feedback_Numerical
from .feedback_create import (
    FeedbackCreate,
    FeedbackCreate_Categorical,
    FeedbackCreate_Numerical,
)
from .feedback_definition_page_public import FeedbackDefinitionPagePublic
from .feedback_object_public import (
    FeedbackObjectPublic,
    FeedbackObjectPublic_Categorical,
    FeedbackObjectPublic_Numerical,
)
from .feedback_public import (
    FeedbackPublic,
    FeedbackPublic_Categorical,
    FeedbackPublic_Numerical,
)
from .feedback_score import FeedbackScore
from .feedback_score_average import FeedbackScoreAverage
from .feedback_score_average_public import FeedbackScoreAveragePublic
from .feedback_score_batch import FeedbackScoreBatch
from .feedback_score_batch_item import FeedbackScoreBatchItem
from .feedback_score_batch_item_source import FeedbackScoreBatchItemSource
from .feedback_score_compare import FeedbackScoreCompare
from .feedback_score_compare_source import FeedbackScoreCompareSource
from .feedback_score_public import FeedbackScorePublic
from .feedback_score_public_source import FeedbackScorePublicSource
from .feedback_score_source import FeedbackScoreSource
from .feedback_update import (
    FeedbackUpdate,
    FeedbackUpdate_Categorical,
    FeedbackUpdate_Numerical,
)
from .json_node import JsonNode
from .json_node_compare import JsonNodeCompare
from .json_node_public import JsonNodePublic
from .json_node_write import JsonNodeWrite
from .numerical_feedback_definition import NumericalFeedbackDefinition
from .numerical_feedback_definition_create import NumericalFeedbackDefinitionCreate
from .numerical_feedback_definition_public import NumericalFeedbackDefinitionPublic
from .numerical_feedback_definition_update import NumericalFeedbackDefinitionUpdate
from .numerical_feedback_detail import NumericalFeedbackDetail
from .numerical_feedback_detail_create import NumericalFeedbackDetailCreate
from .numerical_feedback_detail_public import NumericalFeedbackDetailPublic
from .numerical_feedback_detail_update import NumericalFeedbackDetailUpdate
from .project import Project
from .project_page_public import ProjectPagePublic
from .project_public import ProjectPublic
from .span import Span
from .span_batch import SpanBatch
from .span_page_public import SpanPagePublic
from .span_public import SpanPublic
from .span_public_type import SpanPublicType
from .span_type import SpanType
from .span_write import SpanWrite
from .span_write_type import SpanWriteType
from .trace import Trace
from .trace_batch import TraceBatch
from .trace_page_public import TracePagePublic
from .trace_public import TracePublic
from .trace_write import TraceWrite

__all__ = [
    "CategoricalFeedbackDefinition",
    "CategoricalFeedbackDefinitionCreate",
    "CategoricalFeedbackDefinitionPublic",
    "CategoricalFeedbackDefinitionUpdate",
    "CategoricalFeedbackDetail",
    "CategoricalFeedbackDetailCreate",
    "CategoricalFeedbackDetailPublic",
    "CategoricalFeedbackDetailUpdate",
    "ChunkedOutputJsonNode",
    "ChunkedOutputJsonNodeType",
    "Dataset",
    "DatasetItem",
    "DatasetItemBatch",
    "DatasetItemCompare",
    "DatasetItemCompareSource",
    "DatasetItemPageCompare",
    "DatasetItemPagePublic",
    "DatasetItemPublic",
    "DatasetItemPublicSource",
    "DatasetItemSource",
    "DatasetItemWrite",
    "DatasetItemWriteSource",
    "DatasetPagePublic",
    "DatasetPublic",
    "DeleteFeedbackScore",
    "ErrorMessage",
    "ErrorMessagePublic",
    "Experiment",
    "ExperimentItem",
    "ExperimentItemCompare",
    "ExperimentItemPublic",
    "ExperimentPagePublic",
    "ExperimentPublic",
    "Feedback",
    "FeedbackCreate",
    "FeedbackCreate_Categorical",
    "FeedbackCreate_Numerical",
    "FeedbackDefinitionPagePublic",
    "FeedbackObjectPublic",
    "FeedbackObjectPublic_Categorical",
    "FeedbackObjectPublic_Numerical",
    "FeedbackPublic",
    "FeedbackPublic_Categorical",
    "FeedbackPublic_Numerical",
    "FeedbackScore",
    "FeedbackScoreAverage",
    "FeedbackScoreAveragePublic",
    "FeedbackScoreBatch",
    "FeedbackScoreBatchItem",
    "FeedbackScoreBatchItemSource",
    "FeedbackScoreCompare",
    "FeedbackScoreCompareSource",
    "FeedbackScorePublic",
    "FeedbackScorePublicSource",
    "FeedbackScoreSource",
    "FeedbackUpdate",
    "FeedbackUpdate_Categorical",
    "FeedbackUpdate_Numerical",
    "Feedback_Categorical",
    "Feedback_Numerical",
    "JsonNode",
    "JsonNodeCompare",
    "JsonNodePublic",
    "JsonNodeWrite",
    "NumericalFeedbackDefinition",
    "NumericalFeedbackDefinitionCreate",
    "NumericalFeedbackDefinitionPublic",
    "NumericalFeedbackDefinitionUpdate",
    "NumericalFeedbackDetail",
    "NumericalFeedbackDetailCreate",
    "NumericalFeedbackDetailPublic",
    "NumericalFeedbackDetailUpdate",
    "Project",
    "ProjectPagePublic",
    "ProjectPublic",
    "Span",
    "SpanBatch",
    "SpanPagePublic",
    "SpanPublic",
    "SpanPublicType",
    "SpanType",
    "SpanWrite",
    "SpanWriteType",
    "Trace",
    "TraceBatch",
    "TracePagePublic",
    "TracePublic",
    "TraceWrite",
]



import json
import logging
from typing import Any, List, Optional, Union

from opik import logging_messages
from opik.evaluation.metrics import base_metric, score_result
from opik.evaluation.models import base_model, models_factory

from . import template
from ... import exceptions

LOGGER = logging.getLogger(__name__)


class AnswerRelevance(base_metric.BaseMetric):
    """
    A metric that evaluates the relevance of an answer to a given input using an LLM.

    This metric uses a language model to assess how well the given output (answer)
    addresses the provided input (question) within the given context. It returns a score
    between 0.0 and 1.0, where higher values indicate better answer relevance.

    Args:
        model: The language model to use for evaluation. Can be a string (model name) or a CometBaseModel instance.
        name: The name of the metric. Defaults to "AnswerRelevanceMetric".

    Example:
        >>> from comet_llm_eval.evaluation.metrics import AnswerRelevance
        >>> answer_relevance_metric = AnswerRelevance()
        >>> result = answer_relevance_metric.score("What's the capital of France?", "The capital of France is Paris.", ["France is a country in Europe."])
        >>> print(result.value)
        0.9
        >>> print(result.reason)
        The answer directly addresses the user's query by correctly identifying Paris as the capital of France. ...
    """

    def __init__(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]] = None,
        name: str = "answer_relevance_metric",
        few_shot_examples: Optional[
            List[template.FewShotExampleAnswerRelevance]
        ] = None,
    ):
        super().__init__(
            name=name,
        )

        self._init_model(model)
        if few_shot_examples is None:
            self._few_shot_examples = template.FEW_SHOT_EXAMPLES
        else:
            self._few_shot_examples = few_shot_examples

    def _init_model(
        self, model: Optional[Union[str, base_model.OpikBaseModel]]
    ) -> None:
        if isinstance(model, base_model.OpikBaseModel):
            self._model = model
        else:
            self._model = models_factory.get(model_name=model)

    def score(
        self, input: str, output: str, context: List[str], **ignored_kwargs: Any
    ) -> score_result.ScoreResult:
        """
        Calculate the answer relevance score for the given input-output pair.

        Args:
            input: The input text (question) to be evaluated.
            output: The output text (answer) to be evaluated.
            context: A list of context strings relevant to the input.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object containing the answer relevance score
            (between 0.0 and 1.0) and a reason for the score.
        """
        llm_query = template.generate_query(input=input, output=output, context=context)
        model_output = self._model.generate(input=llm_query)

        return self._parse_model_output(model_output)

    async def ascore(
        self, input: str, output: str, context: List[str], **ignored_kwargs: Any
    ) -> score_result.ScoreResult:
        """
        Asynchronously calculate the answer relevance score for the given input-output pair.

        This method is the asynchronous version of :meth:`score`. For detailed documentation,
        please refer to the :meth:`score` method.

        Args:
            input: The input text (question) to be evaluated.
            output: The output text (answer) to be evaluated.
            context: A list of context strings relevant to the input.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with the answer relevance score and reason.
        """
        llm_query = template.generate_query(input=input, output=output, context=context)
        model_output = await self._model.agenerate(input=llm_query)

        return self._parse_model_output(model_output)

    def _parse_model_output(self, content: str) -> score_result.ScoreResult:
        try:
            dict_content = json.loads(content)
            score: float = dict_content[template.VERDICT_KEY]

            if not (0.0 <= score <= 1.0):
                score = 0.5

            return score_result.ScoreResult(
                name=self.name, value=score, reason=dict_content[template.REASON_KEY]
            )
        except Exception:
            raise exceptions.MetricComputationError(
                logging_messages.ANSWER_RELEVANCE_SCORE_CALC_FAILED
            )



from typing import List
from opik.types import FeedbackScoreDict

from opik.api_objects import opik_client
from . import test_result


def log_scores(
    client: opik_client.Opik, test_results: List[test_result.TestResult]
) -> None:
    all_trace_scores: List[FeedbackScoreDict] = []

    for result in test_results:
        for score_result in result.score_results:
            if score_result.scoring_failed:
                continue

            trace_score = FeedbackScoreDict(
                id=result.test_case.trace_id,
                name=score_result.name,
                value=score_result.value,
                reason=score_result.reason,
            )
            all_trace_scores.append(trace_score)

    client.log_traces_feedback_scores(scores=all_trace_scores)
